import pytest

from unittest.mock import Mock, call

from typing import Any

from pathlib import Path

import shutil

from trees import (
    write_device_tree,
    read_device_tree,
    write_local_tree,
    read_local_tree,
)

from upyt.connection import Connection
from upyt.upy_fs import FilesystemAPI

from upyt.sync import (
    decode_upyt_id,
    encode_upyt_id,
    get_upyt_id,
    enumerate_local_files,
    sync_to_device,
)


class TestUPyTIDCodecs:
    @pytest.mark.parametrize(
        "version, device_id",
        [
            (0, "000000000000"),
            (1, "000000000000"),
            (123, "1234567890aB"),
        ],
    )
    def test_roundtrip(self, version: int, device_id: str) -> None:
        encoded = encode_upyt_id(version, device_id)
        assert decode_upyt_id(encoded) == (version, device_id)

    @pytest.mark.parametrize(
        "value",
        [
            # Missing fields
            b"",
            b"1",
            # Extra fields
            b"1 2 3",
            # Invalid version
            b"nope 000000000000",
            # Invalid device ID (not ASCII)
            b"nope \xFF00000000000",
        ],
    )
    def test_invalid(self, value: bytes) -> None:
        with pytest.raises(ValueError):
            decode_upyt_id(value)


class TestGetUpytId:
    def test_create_file_if_absent(self, fs: FilesystemAPI, dev_tmpdir: str) -> None:
        fs.mkdir(f"{dev_tmpdir}/a")
        fs.mkdir(f"{dev_tmpdir}/b")

        version_a, device_id_a = get_upyt_id(fs, f"{dev_tmpdir}/a")
        version_b, device_id_b = get_upyt_id(fs, f"{dev_tmpdir}/b")

        assert device_id_a != device_id_b

    def test_read_file_if_exists(self, fs: FilesystemAPI, dev_tmpdir: str) -> None:
        version = 123
        device_id = "1234567890aB"
        fs.write_file(f"{dev_tmpdir}/.upyt_id.txt", encode_upyt_id(version, device_id))
        assert get_upyt_id(fs, dev_tmpdir) == (version, device_id)


class TestEnumerateLocalFiles:
    def test_empty(self, tmp_path: Path) -> None:
        assert list(enumerate_local_files(tmp_path)) == []

    def test_empty_with_exclusions(self, tmp_path: Path) -> None:
        assert list(enumerate_local_files(tmp_path, ["foo", "*", "**"])) == []

    def test_flat_files(self, tmp_path: Path) -> None:
        exp = [
            tmp_path / "foo.txt",
            tmp_path / "bar.baz",
        ]
        for file in exp:
            file.touch()

        assert set(enumerate_local_files(tmp_path)) == set(exp)

    def test_nested_files(self, tmp_path: Path) -> None:
        dirs = [
            tmp_path / "foo",
            tmp_path / "foo/bar/",
        ]
        files = [
            tmp_path / "a",
            tmp_path / "foo" / "b",
            tmp_path / "foo" / "bar" / "c",
        ]
        for dir in dirs:
            dir.mkdir()
        for file in files:
            file.touch()

        assert set(enumerate_local_files(tmp_path)) == set(files + dirs)

    def test_exclusions(self, tmp_path: Path) -> None:
        dirs = [
            tmp_path / "foo",
            tmp_path / "foo" / "bar",
            tmp_path / "foo" / "bar" / "exclude_dir",
            tmp_path / "foo" / "bar" / "exclude_txt",
            tmp_path / "foo" / "exclude_in_root",
            tmp_path / "exclude_in_root",
            tmp_path / "exclude_when_directory",
        ]
        files = [
            tmp_path / "a",
            tmp_path / "b.exclude",
            tmp_path / "foo/c.exclude",
            tmp_path / "foo/bar/exclude_dir/d",
            tmp_path / "foo/bar/exclude_txt/e.not_txt",
            tmp_path / "foo/bar/exclude_txt/f.txt",
            tmp_path / "foo/exclude_when_directory",
        ]
        for dir in dirs:
            dir.mkdir()
        for file in files:
            file.touch()

        exclusions = [
            "*.exclude",
            "exclude_dir",
            "exclude_txt/*.txt",
            "/exclude_in_root",
            "exclude_when_directory/",
        ]

        exp_absent = {
            tmp_path / "b.exclude",
            tmp_path / "foo/c.exclude",
            tmp_path / "foo/bar/exclude_dir",
            tmp_path / "foo/bar/exclude_dir/d",
            tmp_path / "foo/bar/exclude_txt/f.txt",
            tmp_path / "exclude_in_root",
            tmp_path / "exclude_when_directory",
        }

        exp = set(files + dirs) - exp_absent

        assert set(enumerate_local_files(tmp_path, exclusions)) == exp


class TestSyncToDevice:
    @pytest.fixture
    def dev_tmpdir_with_id(self, fs: FilesystemAPI, dev_tmpdir: str) -> str:
        """
        A temporary dir initialised with a UPyT ID with a fixed device ID of
        'DEVICEXXXX' and version 500.
        """
        fs.write_file(f"{dev_tmpdir}/.upyt_id.txt", b"500 DEVICEXXXX")
        return dev_tmpdir

    @pytest.mark.parametrize(
        "tree",
        [
            # Empty
            {},
            # Files
            {"a.txt": b"I am a"},
            {"a.txt": b"I am a", "b.txt": b"I am b"},
            # Directories
            {"a": {}},
            {"a": {"b": {}}},
        ],
    )
    def test_create_from_scatch(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
        tree: dict[str, Any],
    ) -> None:
        write_local_tree(tmp_path, tree)
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)

        # Check that the cache exactly reflects the on-device state
        assert read_local_tree(tmp_path) == dict(
            tree,
            **{
                ".upyt_cache": {"DEVICEXXXX": device_tree},
            },
        )

        # Check an incremented device ID was written back
        assert device_tree.pop(".upyt_id.txt") == b"501 DEVICEXXXX"

        # Check the input tree has been written exactly
        assert device_tree == tree

    def test_exclusions(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        write_local_tree(
            tmp_path,
            {
                "a": b"file a",
                "b": b"file b",
                "c.exclude": b"file c (excluded)",
            },
        )
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id, exclude=["*.exclude"])
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)

        # Check the input tree has been written exactly
        device_tree.pop(".upyt_id.txt")
        assert device_tree == {
            "a": b"file a",
            "b": b"file b",
        }

    def test_add_then_remove(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        tree = {
            "a": {"subdir": {}},
            "b": {"subfile": b"I am a subfile"},
            "c": b"top level file",
        }

        write_local_tree(tmp_path, tree)
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)

        # Check that the cache exactly reflects the on-device state
        assert read_local_tree(tmp_path) == dict(
            tree,
            **{
                ".upyt_cache": {"DEVICEXXXX": device_tree},
            },
        )

        # Check an incremented device ID was written back
        assert device_tree.pop(".upyt_id.txt") == b"501 DEVICEXXXX"

        # Check the input tree has been written exactly
        assert device_tree == tree

        # Remove files
        shutil.rmtree(tmp_path / "a")
        shutil.rmtree(tmp_path / "b")
        (tmp_path / "c").unlink()

        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)

        # Check that the cache reflects the values we just uploaded (i.e.
        # nothing!)
        assert read_local_tree(tmp_path) == {
            ".upyt_cache": {
                "DEVICEXXXX": {
                    ".upyt_id.txt": b"502 DEVICEXXXX",
                }
            },
        }

        # Check an incremented device ID was written back
        assert device_tree.pop(".upyt_id.txt") == b"502 DEVICEXXXX"

        # The extra files should still remain on the device
        assert device_tree == tree

    def test_switch_between_file_and_dir(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        tree = {
            "was_dir": {},
            "was_file": b"I am a file for now",
        }
        write_local_tree(tmp_path, tree)
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        # Check as expected
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        device_tree.pop(".upyt_id.txt")
        assert device_tree == tree

        new_tree = {
            "was_dir": b"I am a file for now",
            "was_file": {},
        }
        write_local_tree(tmp_path, new_tree)
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        # Check types changed
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        device_tree.pop(".upyt_id.txt")
        assert device_tree == new_tree

    def test_update_content(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        write_local_tree(tmp_path, {"file": b"Foo" + (b"X" * 1024)})
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        write_local_tree(tmp_path, {"file": b"Bar" + (b"X" * 1024)})
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        assert device_tree.pop("file") == (b"Bar" + (b"X" * 1024))

    def test_update_content_catch_stale_cache_file(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        # Deliberately Corrupt the cache add a file which isn't yet on the
        # device (i.e. which hypothetically was deleted earlier on)
        cache_dir = tmp_path / ".upyt_cache" / "DEVICEXXXX"
        cache_dir.mkdir(parents=True)
        (cache_dir / ".upyt_id.txt").write_bytes(b"400 DEVICEXXXX")
        (cache_dir / "file").write_bytes(b"Something")

        # Should still overwrite file even though the (out-of-date) cache
        # suggests it is already on the device
        write_local_tree(tmp_path, {"file": b"Something"})
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        assert device_tree.pop("file") == b"Something"

    def test_force_enumerate_files(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        # Deliberately create a corrupt cache which would suggest a file is on
        # the device but isn't in reality to verify the force-mechanism works
        cache_dir = tmp_path / ".upyt_cache" / "DEVICEXXXX"
        cache_dir.mkdir(parents=True)
        (cache_dir / ".upyt_id.txt").write_bytes(b"500 DEVICEXXXX")
        (cache_dir / "file").write_bytes(b"Something")

        write_local_tree(tmp_path, {"file": b"Something"})

        # No force, no file
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)
        assert "file" not in read_device_tree(fs, dev_tmpdir_with_id)

        # Force leads to file existing
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id, force_enumerate_files=True)
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        assert device_tree.pop("file") == b"Something"

    def test_update_content_catch_stale_cache_value(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        write_local_tree(tmp_path, {"file": b"Foo" + (b"X" * 1024)})
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        # Deliberately Corrupt the cache (different version and different file
        # content)
        cache_dir = tmp_path / ".upyt_cache" / "DEVICEXXXX"
        (cache_dir / ".upyt_id.txt").write_bytes(b"500 DEVICEXXXX")
        (cache_dir / "file").write_bytes(b"Something else")

        # Should silently catch issue and fall back on writing file directly
        # leading to correct outcome
        write_local_tree(tmp_path, {"file": b"Something different"})
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)

        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        assert device_tree.pop("file") == (b"Something different")

    def test_force_safe_update(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        # Deliberately create a corrupt cache which would suggest a file is on
        # the device with a particular value which doesn't match reality.
        cache_dir = tmp_path / ".upyt_cache" / "DEVICEXXXX"
        cache_dir.mkdir(parents=True)
        (cache_dir / ".upyt_id.txt").write_bytes(b"500 DEVICEXXXX")
        (cache_dir / "file").write_bytes(b"Something else")

        fs.write_file(f"{dev_tmpdir_with_id}/file", b"Entirely unrelated")

        write_local_tree(tmp_path, {"file": b"Something different"})

        # Without safe mode, the corrupt cache should lead to a corrupt file
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id)
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        assert device_tree.pop("file") == (b"Entirely udifferent")  # Corrupt!

        # With safe mode, the corrupt cache should be caught and the correct
        # file written
        (cache_dir / "file").write_bytes(b"Something else")
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id, force_safe_update=True)
        device_tree = read_device_tree(fs, dev_tmpdir_with_id)
        assert isinstance(device_tree, dict)
        assert device_tree.pop("file") == (b"Something different")

    def test_progress_reporting(
        self,
        fs: FilesystemAPI,
        dev_tmpdir_with_id: str,
        tmp_path: Path,
    ) -> None:
        write_local_tree(tmp_path, {"a": b"I am 'a'!", "b": b"I am 'b'!"})

        mock = Mock()
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id, progress_callback=mock)

        # Should have notified on every change
        assert len(mock.mock_calls) == 2
        mock.assert_has_calls(
            [
                call(
                    Path(name),
                    {Path("a"), Path("b")},
                    {Path("a"), Path("b")},
                )
                for name in ["a", "b"]
            ]
        )

        # Minor change
        write_local_tree(tmp_path, {"b": b"I am a changed 'b'!", "c": b"I am 'c'"})

        mock = Mock()
        sync_to_device(fs, tmp_path, dev_tmpdir_with_id, progress_callback=mock)

        # Should have notified on only differences
        assert len(mock.mock_calls) == 2
        mock.assert_has_calls(
            [
                call(
                    Path(name),
                    {Path("b"), Path("c")},
                    {Path("a"), Path("b"), Path("c")},
                )
                for name in ["b", "c"]
            ]
        )



import pytest
from unittest.mock import Mock

from typing import Any, Callable, Iterable

import time

from threading import Thread
from queue import Queue
from functools import partial
from concurrent.futures import ThreadPoolExecutor

import websockets.sync.server as ws_server

from upyt.connection import Connection, WebReplAuthenticationError


class TestConnectionFromSpecification:
    @pytest.fixture
    def mock_serial_connection(self, monkeypatch: Any) -> Mock:
        # Mock out SerialConnection
        import upyt.connection

        mock = Mock()
        monkeypatch.setattr(upyt.connection, "SerialConnection", mock)

        mock.return_value = mock

        return mock

    @pytest.mark.parametrize("name", ["COM2", "/dev/ttyACM3"])
    def test_serial_default_baudrate(
        self, name: str, mock_serial_connection: Mock
    ) -> None:
        out = Connection.from_specification(name)

        assert out is mock_serial_connection
        mock_serial_connection.assert_called_once_with(name, 9600)

    @pytest.mark.parametrize(
        "name", ["COM2", "/dev/ttyACM3", "/dev/ttyACM:with:colons:in"]
    )
    def test_serial_custom_baudrate(
        self, name: str, mock_serial_connection: Mock
    ) -> None:
        out = Connection.from_specification(f"{name}:115200")

        assert out is mock_serial_connection
        mock_serial_connection.assert_called_once_with(name, 115200)

    @pytest.fixture
    def mock_web_repl_connection(self, monkeypatch: Any) -> Mock:
        # Mock out WebReplConnection
        import upyt.connection

        mock = Mock()
        monkeypatch.setattr(upyt.connection, "WebReplConnection", mock)

        mock.return_value = mock

        return mock

    def test_web_repl_valid(self, mock_web_repl_connection: Mock) -> None:
        out = Connection.from_specification(f"ws://foo.bar/baz?pw1234")
        mock_web_repl_connection.assert_called_once_with("ws://foo.bar/baz", "pw1234")

    def test_web_repl_no_password(self, mock_web_repl_connection: Mock) -> None:
        with pytest.raises(ValueError):
            Connection.from_specification(f"ws://foo.bar/baz")


# A (rx_fn, tx_fn, ws_url) tuple
ServerFixture = tuple[Callable[[], str], Callable[[str], None], str]

# A (rx_fn, tx_fn, Connection) tuple
AuthedServerFixture = tuple[Callable[[], str], Callable[[str], None], Connection]


class TestWebReplConnection:
    @pytest.fixture
    def server(self) -> Iterable[ServerFixture]:
        """
        Create a local websocket server on a random port.

        Fixture returns a (recv_message, send_message, url) tuple.
        """
        rx_queue: Queue[str] = Queue()
        tx_queue: Queue[str] = Queue()
        connections: list[ws_server.ServerConnection] = []

        def rx_loop(connection: ws_server.ServerConnection):
            for message in connection:
                assert isinstance(message, str)
                rx_queue.put(message)

        def tx_loop(connection: ws_server.ServerConnection):
            while message := tx_queue.get():
                connection.send(message)

        def on_connect(connection: ws_server.ServerConnection) -> None:
            # NB: Only one connection supported
            if connections:
                connection.close()
                print("Error: Tried to open multiple connections")
            connections.append(connection)

            tx_thread = Thread(
                target=tx_loop, args=(connection,), daemon=True, name="tx"
            )
            tx_thread.start()

            rx_loop(connection)

            tx_thread.join()

        server = ws_server.serve(on_connect, host="127.0.0.1", port=0)
        assigned_port = server.socket.getsockname()[1]

        server_thread = Thread(target=server.serve_forever, name="server")
        try:
            server_thread.start()
            yield (
                partial(rx_queue.get, timeout=1),
                tx_queue.put,
                f"ws://127.0.0.1:{assigned_port}/",
            )
        finally:
            server.shutdown()
            tx_queue.put("")  # Terminate the TX thread
            for connection in connections:
                connection.close()
            server_thread.join()

    def test_authentication(self, server: ServerFixture) -> None:
        rx, tx, url = server

        fut = ThreadPoolExecutor().submit(Connection.from_specification, f"{url}?pw123")

        tx("Password: ")
        assert rx() == "pw123\n"
        tx("\r\nWebREPL connected\r\n>>> ")

        conn = fut.result(1)

        # Check the connection then becomes a bidirectional channel
        conn.write(b"Hello!")
        assert rx() == "Hello!"
        tx("World!")
        assert conn.read_until(b"!") == b"World!"

    def test_password_incorrect(self, server: ServerFixture) -> None:
        rx, tx, url = server

        fut = ThreadPoolExecutor().submit(Connection.from_specification, f"{url}?pw123")

        tx("Password: ")
        assert rx() == "pw123\n"
        tx("\r\nInvalid password\r\n")

        with pytest.raises(WebReplAuthenticationError):
            fut.result(1)

    @pytest.fixture
    def authed_server(self, server: ServerFixture) -> AuthedServerFixture:
        """
        Like the server fixture but returns a (rx, tx, conn) tuple which is
        ready-to-use.
        """
        rx, tx, url = server

        fut = ThreadPoolExecutor().submit(Connection.from_specification, f"{url}?pw123")

        tx("Password: ")
        assert rx() == "pw123\n"
        tx("\r\nWebREPL connected\r\n>>> ")

        conn = fut.result(1)

        return (rx, tx, conn)

    def test_read(self, authed_server: AuthedServerFixture) -> None:
        rx, tx, conn = authed_server

        tx("123")
        tx("456")

        # Can read subset of message
        assert conn.read(2) == b"12"

        # Can read accross message boundaries
        assert conn.read(2) == b"34"

        # Timeout works
        with conn.timeout_override(0.1):
            before = time.monotonic()
            assert conn.read(4) == b"56"
            after = time.monotonic()
            assert after - before >= conn.timeout

    def test_read_until(self, authed_server: AuthedServerFixture) -> None:
        rx, tx, conn = authed_server

        tx("123")
        tx("456")

        # Can read subset of message
        assert conn.read_until(b"2") == b"12"

        # Can read accross message boundaries
        assert conn.read_until(b"4") == b"34"

        # Timeout works
        with conn.timeout_override(0.1):
            before = time.monotonic()
            assert conn.read_until(b"7") == b"56"
            after = time.monotonic()
            assert after - before >= conn.timeout

    def test_read_buffered(self, authed_server: AuthedServerFixture) -> None:
        rx, tx, conn = authed_server

        tx("123")
        time.sleep(0.1)
        assert conn.read_buffered() == b"123"

    def test_write(self, authed_server: AuthedServerFixture) -> None:
        rx, tx, conn = authed_server

        # NB: We can't test writing non-UTF-8 bytestreams since the websocket
        # server we're using is actually conformant(!)
        conn.write(b"123")
        conn.write(b"456")

        assert rx() == "123"
        assert rx() == "456"



import pytest

from unittest.mock import Mock
from typing import Iterator
from io import StringIO
from itertools import islice
from pathlib import Path

from upyt.read_proxy import (
    replace,
    match,
    ReadProxy,
)


class TestReplace:
    def test_replacement(self) -> None:
        seq = iter("abc12312X")
        out = replace("123", "<1-2-3>")(seq)
        for got, exp in zip(out, "abc<1-2-3>12X"):
            assert got == exp

    def test_retain_while_uncertain(self) -> None:
        seq = iter(list("abc12") + [None] + list("3"))
        out = replace("123", "<1-2-3>")(seq)
        for got, exp in zip(out, list("abc") + [None] + list("<1-2-3>")):
            assert got == exp


class TestCallOnMatch:
    def test_called(self) -> None:
        seq = "abc12312X"
        mock = Mock()
        out = match("123", mock)(seq)

        # Should not be called initially
        for exp, got in zip("abc", out):
            assert not mock.called
            assert got == exp

        # When we encounter the matched sequence we should be called
        for exp, got in zip("123", out):
            assert len(mock.mock_calls) == 1
            assert got == exp

        # And whatever follows should come out afterwards (with no further
        # calls)
        for exp, got in zip("12X", out):
            assert len(mock.mock_calls) == 1
            assert got == exp

    def test_retain_while_uncertain(self) -> None:
        seq = iter(list("abc12") + [None] + list("3"))
        mock = Mock()
        out = match("123", mock)(seq)

        for exp, got in zip(list("abc") + [None], out):
            assert not mock.called
            assert got == exp

        for exp, got in zip(list("123"), out):
            assert len(mock.mock_calls) == 1
            assert got == exp

    def test_breakout_with_exception(self) -> None:
        seq = iter("abc123456")
        mock = Mock(side_effect=NotImplementedError)
        out = match("123", mock)(seq)

        for exp, got in zip("abc", out):
            assert got == exp

        # When called, we should hit an exception as expected
        with pytest.raises(NotImplementedError):
            next(out)

        # The input should have been consumed exactly up to the end of the
        # matched sequence -- and no more!
        assert "".join(seq) == "456"


class TestReadProxy:
    def test_read_whole(self) -> None:
        f = StringIO("ABC")
        p = ReadProxy(f, [])

        assert p.read() == "ABC"
        assert p.read() == ""

    def test_read_partial(self) -> None:
        f = StringIO("ABC")
        p = ReadProxy(f, [])

        assert p.read(1) == "A"
        assert p.read(2) == "BC"
        assert p.read(3) == ""

    def test_fileno(self, tmp_path: Path) -> None:
        filename = tmp_path / "file.txt"
        filename.write_text("ABC")
        with filename.open("r") as f:
            p = ReadProxy(f, [])
            assert p.fileno() == f.fileno()

    def test_filters(self, tmp_path: Path) -> None:
        f = StringIO("ABC12312X")
        p = ReadProxy(
            f,
            [
                replace("123", "<1-2-3>"),
                replace("<", "<<<"),
                replace(">", ">>>"),
            ],
        )

        assert p.read() == "ABC<<<1-2-3>>>12X"



from argparse import ArgumentParser

import os

from importlib import import_module


def main(argv: list[str] | None = None) -> None:
    parser = ArgumentParser(
        description="""
            A multi-tool for programming and interacting with MicroPython
            devices.
        """
    )
    parser.add_argument(
        "--device",
        "-d",
        default=os.getenv("UPYT_DEVICE"),
        help="""
            Device to connect to. For example `/dev/ttyACM0` or
            `/dev/ttyACM0:115200`. Defaults to value of UPYT_DEVICE environment
            variable, if set, required otherwise.
        """,
    )

    subparsers = parser.add_subparsers(title="commands", required=True)

    subcommands = [
        (["terminal", "term", "t"], "a serial terminal for MicroPython."),
        ("sync", "efficiently synchronise a directory to the device"),
        ("reset", "perform a soft-reset"),
        (["interrupt", "int"], "interrupt and return to the REPL"),
        ("ls", "list files and directories"),
        ("mkdir", "create a directory"),
        ("rm", "remove files and directories"),
        ("cat", "read (and concatenate) files"),
        ("cp", "copy files to and from the device"),
    ]

    for command, help_text in subcommands:
        if isinstance(command, str):
            command = [command]

        command_module = import_module(f"upyt.cli.{command[0]}")
        subparser = subparsers.add_parser(
            command[0],
            help=help_text,
            description=command_module.__doc__,
            aliases=command[1:],
        )
        command_module.add_arguments(subparser)
        subparser.set_defaults(cmd=command_module.main)

    args = parser.parse_args(argv)

    if args.device is None:
        parser.error("--device is required if UPYT_DEVICE is not set")

    args.cmd(args)


if __name__ == "__main__":
    main()



from typing import Optional

import pytest

import time

from upyt.connection import Connection

from unittest.mock import Mock

from textwrap import dedent

from contextlib import contextmanager

from upyt.upy_repl import (
    MicroPythonReplError,
    expect,
    expect_endswith,
    interrupt_and_enter_repl,
    paste_exec,
    raw_mode,
    raw_paste_exec,
    soft_reset_directly_into_repl,
)


class TestExpect:
    def test_matches(self) -> None:
        ser = Mock()
        ser.read.return_value = b"yes"

        assert expect(ser, b"yes") == b"yes"

    def test_no_match(self) -> None:
        ser = Mock()
        ser.read.return_value = b"no"

        with pytest.raises(MicroPythonReplError) as exc_info:
            expect(ser, b"yes")

        assert exc_info.value.args[0] == b"no"


class TestExpectEndswIth:
    def test_matches(self) -> None:
        ser = Mock()
        ser.read_until.return_value = b"well...yes"

        assert expect_endswith(ser, b"yes") == b"well...yes"

    def test_no_match(self) -> None:
        ser = Mock()
        ser.read_until.return_value = b"well...no"

        with pytest.raises(MicroPythonReplError) as exc_info:
            expect_endswith(ser, b"yes")

        assert exc_info.value.args[0] == b"well...no"


class TestInterruptAndEnterRepl:
    def test_no_contention_next_prompt_unread(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        # Confirm the REPL is working and that we've gone back to a prompt
        ser.write(b"print('hello')\r")
        assert ser.read_until(b"\r\nhello") == (
            b"print('hello')\r\n" b"hello"  # Echo  # Output
        )

        # NB: newline and prompt (>>>) still left in buffer
        assert interrupt_and_enter_repl(ser) == b"\r\n>>> "

    def test_no_contention_next_prompt_read(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        # Confirm the REPL is working and that we've gone back to a prompt
        ser.write(b"print('hello')\r")
        assert ser.read_until(b"hello\r\n>>> ") == (
            b"print('hello')\r\n" b"hello\r\n" b">>> "  # Echo  # Output  # Next prompt
        )

        # NB: nothing left in buffer
        assert interrupt_and_enter_repl(ser) == b""

    def test_interrupt_process(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        before = time.time()

        ser.write(b"import time; time.sleep(30)\r")

        # XXX: Due to an unknown issue with the ACM driver/pyserial/something
        # else(?), if we do not wait for the echo to come back, the attempt to
        # write a ^C character in interrupt_and_enter_repl() will be blocked
        # until the end of the sleep for reasons unknown.
        expect(ser, b"import time; time.sleep(30)\r\n")

        time.sleep(0.1)  # Ensure started before we send ctrl+c

        # Ensure we interrupt the sleep
        out = interrupt_and_enter_repl(ser)
        print(out.decode("utf-8"))
        after = time.time()

        # Make sure our ^C worked (and that we didn't end up blocked waiting
        # for the printout
        assert after - before < ser.timeout * 1.5

        assert out == (
            b"Traceback (most recent call last):\r\n"
            b'  File "<stdin>", line 1, in <module>\r\n'
            b"KeyboardInterrupt: "
        )

    def test_interrupt_process_custom_message(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        before = time.time()

        # XXX: Due to an unknown issue with the ACM driver/pyserial/something
        # else(?), if we do not wait for the echo to come back, the attempt to
        # write a ^C character in interrupt_and_enter_repl() will be blocked
        # until the end of the sleep for reasons unknown.
        ser.write(b"import time\r")
        expect(ser, b"import time\r\n")
        ser.write(b"try: time.sleep(30)\r")
        expect(ser, b">>> try: time.sleep(30)\r\n")
        ser.write(b"except: print('stopped\\ndead')\r")
        expect(ser, b"... except: print('stopped\\ndead')\r\n")
        ser.write(b"\r")
        expect(ser, b"... \r\n")

        time.sleep(0.1)  # Ensure started before we send ctrl+c

        # Ensure we interrupt the sleep
        out = interrupt_and_enter_repl(ser)
        after = time.time()

        # Make sure our ^C worked (and that we didn't end up blocked waiting
        # for the printout
        assert after - before < ser.timeout * 1.5

        assert out == (b"stopped\r\n" b"dead")

    def test_double_interrupt_process(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        before = time.time()

        ser.write(b"import time\r")
        expect(ser, b"import time\r\n>>> ")
        ser.write(b"try: time.sleep(30)\r")
        expect(ser, b"try: time.sleep(30)\r\n... ")
        ser.write(b"finally: time.sleep(20)\r")
        expect(ser, b"finally: time.sleep(20)\r\n... ")
        ser.write(b"\r")
        expect(ser, b"\r\n")

        time.sleep(0.1)  # Ensure started before we send ctrl+c

        # XXX: Due to an unknown issue with the ACM driver/pyserial/something
        # else(?), if we do not wait for the echo to come back, the attempt to
        # write a ^C character in interrupt_and_enter_repl() will be blocked
        # until the end of the sleep for reasons unknown.

        # Ensure we interrupt the sleep
        out = interrupt_and_enter_repl(ser)
        after = time.time()

        # Make sure our ^C worked (and that we didn't end up blocked waiting
        # for the printout
        delta = after - before
        assert delta < ser.timeout * 2.5

        assert out == (
            b"Traceback (most recent call last):\r\n"
            b'  File "<stdin>", line 2, in <module>\r\n'
            b"KeyboardInterrupt: "
        )


class TestPasteExec:
    def test_basic_functionality(self, ser: Connection) -> None:
        paste_exec(
            ser,
            (
                "print('Hello')\n"  # NL separated
                "print('World')\r\n"  # CRNL separated
                "print('£1.23')"  # Unicode
            ),
        )
        expect(ser, "Hello\r\nWorld\r\n£1.23\r\n".encode("utf-8"))  # Output
        expect(ser, b">>> ")  # New prompt

    def test_flow_control(self, ser: Connection) -> None:
        # The following snippet is long enough that not waiting for the remote
        # device to handle it will inevetably result in it loosing some bytes
        paste_exec(ser, f"print({' + '.join(map(str, range(30)))})\n" * 30)
        expect(ser, (str(sum(range(30))).encode("utf-8") + b"\r\n") * 30)  # Output
        expect(ser, b">>> ")  # New prompt


def test_raw_mode(ser: Connection) -> None:
    with raw_mode(ser):
        # Check we're actually in raw mode
        ser.write(b"print('hello')" b"\x04")  # (Ctrl+D)
        expect(ser, b"OKhello\r\n\x04\x04")

    # Check returned to non-raw mode
    ser.write(b"print('hello')\r")
    expect(ser, b"print('hello')\r\n")  # Echo
    expect(ser, b"hello\r\n")
    expect(ser, b">>> ")  # New prompt


class TestRawPasteExec:
    def test_basic_functionality_no_exception(self, ser: Connection) -> None:
        with raw_mode(ser):
            out, err = raw_paste_exec(
                ser,
                dedent(
                    r"""
                        print("hello\nworld!")
                        print("goodbye!")
                    """
                ),
            )
            assert err == ""
            assert out == ("hello\r\n" "world!\r\n" "goodbye!\r\n")

    def test_basic_functionality_with_exception(self, ser: Connection) -> None:
        with raw_mode(ser):
            out, err = raw_paste_exec(
                ser,
                dedent(
                    r"""
                        print("success")
                        raise Exception("Failure!")
                    """
                ),
            )
            assert out == "success\r\n"
            assert err == (
                "Traceback (most recent call last):\r\n"
                '  File "<stdin>", line 3, in <module>\r\n'
                "Exception: Failure!\r\n"
            )

    def test_input_flow_control_required(self, ser: Connection) -> None:
        with raw_mode(ser):
            out, err = raw_paste_exec(
                ser, f"print({'+'.join('1' for _ in range(1000))})"
            )
            assert err == ""
            assert out == "1000\r\n"

    def test_input_flow_control_size_multiple_of_window(self, ser: Connection) -> None:
        # XXX: Hard coded based on what the Pi Pico defaults to...
        window_size = 128

        with raw_mode(ser):
            out, err = raw_paste_exec(ser, f"print(123)".ljust(window_size * 8, "#"))
            assert err == ""
            assert out == "123\r\n"

    def test_disallow_ctrl_d_in_source(self, ser: Connection) -> None:
        with raw_mode(ser):
            with pytest.raises(ValueError):
                raw_paste_exec(ser, "oh noes\x04")


def quick_and_dirty_write_file(
    ser: Connection, file: str, value: Optional[str]
) -> None:
    interrupt_and_enter_repl(ser)

    with raw_mode(ser):
        if value is not None:
            assert (
                raw_paste_exec(
                    ser,
                    dedent(
                        f"""
                        with open({file!r}, "w") as f:
                            f.write({value!r})
                    """
                    ).strip(),
                )
                == ("", "")
            )
        else:
            assert (
                raw_paste_exec(
                    ser,
                    dedent(
                        f"""
                        import os
                        try:
                            os.remove({file!r})
                        except OSError:
                            pass
                    """
                    ).strip(),
                )
                == ("", "")
            )


def quick_and_dirty_read_file(ser: Connection, file: str) -> Optional[str]:
    interrupt_and_enter_repl(ser)

    with raw_mode(ser):
        out, err = raw_paste_exec(
            ser,
            dedent(
                f"""
                    try:
                        print(repr(open({file!r}, "r").read()))
                    except OSError:
                        print("None")
                """
            ).strip(),
        )

    assert err == ""
    return eval(out)


@contextmanager
def quick_and_dirty_override_file(ser: Connection, file: str, content: Optional[str]):
    original = quick_and_dirty_read_file(ser, file)
    quick_and_dirty_write_file(ser, file, content)
    try:
        yield
    finally:
        quick_and_dirty_write_file(ser, file, original)


@pytest.mark.parametrize(
    "original, replacement",
    [
        ("foo", "bar"),
        ("foo", None),
        (None, "foo"),
        (None, None),
    ],
)
def test_override_file(
    ser: Connection,
    original: Optional[str],
    replacement: Optional[str],
) -> None:
    quick_and_dirty_write_file(ser, "file.txt", original)

    with quick_and_dirty_override_file(ser, "file.txt", replacement):
        assert quick_and_dirty_read_file(ser, "file.txt") == replacement

    assert quick_and_dirty_read_file(ser, "file.txt") == original


class TestSoftResetDirectlyIntoRepl:
    def test_interrupt_existing_process(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        # Set the thing off sleeping
        ser.write(b"import time; time.sleep(3)\r")
        expect(ser, b"import time; time.sleep(3)\r\n")

        assert soft_reset_directly_into_repl(ser) == ""

    def test_skips_main_py(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        with quick_and_dirty_override_file(
            ser, "main.py", "import time; time.sleep(3)"
        ):
            # Should have no boot message
            assert soft_reset_directly_into_repl(ser) == ""

            # Should already be in the shell
            ser.write(b"\r")
            expect(ser, b"\r\n>>> ")

    def test_captures_boot_py_output(self, ser: Connection) -> None:
        interrupt_and_enter_repl(ser)

        with quick_and_dirty_override_file(ser, "boot.py", "print('booted!')"):
            # Should have no boot message
            assert soft_reset_directly_into_repl(ser) == "booted!\r\n"



"""
For internal use.

This module contains machinery for producing 'read proxies' which sit between
stdin and the :py:mod:`upyt.upy_terminal` terminal to handle things like
keyboard shortcuts (e.g. to quit the terminal) and intercept bracketed paste
sequences.

The :py:class:`ReadProxy` class takes a file and a
seuqnce of stream filters to apply to the stream.

A stream filter is a factory which takes an iterator-of-characters as an
argument and itself returns an iterator-of-characters following whatever
filtering/processing is needed. Iterators may generate a None (rather than a
character) if they are waiting for further input.
"""


from typing import cast, TextIO, Iterable, Iterator, Callable, AnyStr, BinaryIO, Any

from itertools import takewhile, islice


StreamFilter = Callable[[Iterable[str | None]], Iterator[str | None]]


def replace(find: str, replace: str) -> StreamFilter:
    """
    Return a stream filter which replaces all instances of 'find' with
    'replace'. Partial matches are buffered until they are determined to
    unambiguously match or not.
    """

    def replacer(sequence: Iterable[str | None]) -> Iterator[str | None]:
        # We may need to buffer the input to decide if what we've read matches the
        # find sequence
        input_buffer = ""

        for char in sequence:
            if char is None:
                yield None
            else:
                input_buffer += char
                if find == input_buffer:
                    # Found a match, produce its replacement
                    for out in replace:
                        yield out
                    input_buffer = ""
                elif not find.startswith(input_buffer):
                    # What we've buffered so far can't possibly be a match so
                    # return it as-is.
                    for out in input_buffer:
                        yield out
                    input_buffer = ""
                else:
                    # Otherwise, we've buffered the start of what might be a match,
                    # lets get more and see...
                    pass

    return replacer


def match(find: str, callback: Callable[[], None]) -> StreamFilter:
    """
    Return a stream filter which calls the provided callback whenever the
    'find' string is encountered. Like :py:func:`replace`, buffers values which
    may potentially match 'find' until the ambiguity is resolved.

    Note that the callback is called as soon as the find string has been
    encountered but prior to the string being forwarded. This means if the
    callback throws an exception, the matched characters will never be emitted.
    """

    def matcher(sequence: Iterable[str | None]) -> Iterator[str | None]:
        # We may need to buffer the input to decide if what we've read matches the
        # find sequence
        input_buffer = ""

        for char in sequence:
            if char is None:
                yield None
            else:
                input_buffer += char

                if find == input_buffer:
                    # Found what we're looking for
                    callback()

                if find == input_buffer or not find.startswith(input_buffer):
                    # Either found what we were looking for, or definitely don't
                    # have a partial match so just past it on unchanged.
                    for out in input_buffer:
                        yield out
                    input_buffer = ""

    return matcher


class ReadProxy(TextIO):
    def __init__(
        self,
        readable: TextIO,
        stream_filters: list[StreamFilter],
    ) -> None:
        """
        Create a read proxy which proxies reads from 'readable' via the series
        of stream filters provided. Filters are applied in the order
        they're provided.
        """
        self._readable = readable

        self._iterator = self._read_iter()
        for stream_filter in stream_filters:
            self._iterator = stream_filter(self._iterator)

    def _read_iter(self) -> Iterator[str | None]:
        """
        Wrap the internal file object in an iterator of characters (or None
        when no data is available for reading).
        """
        while True:
            char = self._readable.read(1)
            yield char or None

    def read(self, length=None) -> str:
        iterator = cast(
            Iterable[str],
            takewhile(lambda b: b is not None, self._iterator),
        )
        if length is not None:
            iterator = islice(iterator, length)

        return "".join(iterator)

    def fileno(self) -> int:
        return self._readable.fileno()

    # This class is only a bare-bones look-alike for a readable file for
    # internal use so all of the following are not implemented...

    @property
    def mode(self) -> str:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()

    @property
    def closed(self) -> bool:
        raise NotImplementedError()

    def flush(self) -> None:
        raise NotImplementedError()

    def isatty(self) -> bool:
        raise NotImplementedError()

    def readable(self) -> bool:
        raise NotImplementedError()

    def readline(self, limit: int = -1) -> AnyStr:
        raise NotImplementedError()

    def readlines(self, hint: int = -1) -> list[AnyStr]:
        raise NotImplementedError()

    def seek(self, offset: int, whence: int = 0) -> int:
        raise NotImplementedError()

    def seekable(self) -> bool:
        raise NotImplementedError()

    def tell(self) -> int:
        raise NotImplementedError()

    def truncate(self, size: int | None = None) -> int:
        raise NotImplementedError()

    def writable(self) -> bool:
        raise NotImplementedError()

    def write(self, s: AnyStr) -> int:
        raise NotImplementedError()

    def writelines(self, lines: Iterable[AnyStr]) -> None:
        raise NotImplementedError()

    def __exit__(self, type, value, traceback) -> None:
        raise NotImplementedError()

    @property
    def buffer(self) -> BinaryIO:
        raise NotImplementedError()

    @property
    def encoding(self) -> str:
        raise NotImplementedError()

    @property
    def errors(self) -> str | None:
        raise NotImplementedError()

    @property
    def line_buffering(self) -> bool:
        raise NotImplementedError()

    @property
    def newlines(self) -> Any:
        raise NotImplementedError()

    def __enter__(self) -> "TextIO":
        raise NotImplementedError()

    def __iter__(self) -> Iterator[str]:
        raise NotImplementedError()

    def __next__(self) -> str:
        raise NotImplementedError()



"""
Interrupt a MicroPython device, returning the interpreter to the REPL.
"""

import sys

from argparse import ArgumentParser, Namespace

from upyt.connection import Connection
from upyt.upy_repl import interrupt_and_enter_repl


def add_arguments(parser: ArgumentParser) -> None:
    pass


def main(args: Namespace):
    with Connection.from_specification(args.device) as conn:
        interrupt_and_enter_repl(conn)



"""
A serial terminal for MicroPython.
"""

from argparse import ArgumentParser, Namespace

from upyt.connection import Connection
from upyt.upy_terminal import serial_terminal, GREY, RESET


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="""
            If given, don't print instructions for exiting the terminal on
            startup.
        """,
    )
    parser.add_argument(
        "--no-automatic-paste-mode",
        "-P",
        action="store_true",
        help="""
            If given, disable the automatic use of paste mode when multiple
            lines of text are pasted into the terminal.
        """,
    )
    parser.add_argument(
        "--no-emulate-ctrl-l",
        "-L",
        action="store_true",
        help="""
            If given, disable emulation of the Ctrl+L (clear terminal) keyboard
            shortcut.
        """,
    )


def terminal(
    conn: Connection,
    args: Namespace,
    extra_exit_on: list[str] = [],
    extra_help: str = "",
    show_help: bool = True,
) -> str:
    """
    Run the terminal, returning the exit sequence used to exit it.
    """
    if not args.quiet and show_help:
        print(f"{GREY}Press Ctrl+] to exit. {extra_help}{RESET}")

    exit_seq = serial_terminal(
        conn,
        exit_on=["\x1d"] + extra_exit_on,  # Ctrl+]
        automatic_paste_mode=not args.no_automatic_paste_mode,
        emulate_ctrl_l=not args.no_emulate_ctrl_l,
    )
    print()  # Move to new line on exit
    return exit_seq


def main(args: Namespace):
    with Connection.from_specification(args.device) as conn:
        terminal(conn, args)



"""
Higher-level utilities for (relatively efficiently) manipulating the filesystem
on a MicroPython device via its repl.
"""

from typing import Iterable, Iterator, Optional, Callable

from enum import Enum

import os

from contextlib import contextmanager

from textwrap import dedent

from binascii import hexlify, unhexlify

from hashlib import sha256

from difflib import SequenceMatcher

from upyt.upy_repl import (
    raw_mode,
    raw_paste_exec,
)

from upyt.connection import Connection


@contextmanager
def upy_filesystem(conn: Connection) -> Iterator["FilesystemAPI"]:
    with raw_mode(conn):
        yield FilesystemAPI(conn)


def traceback_to_oserror(traceback: str) -> None:
    """
    Convert a traceback involving an OSError into a raised OSError, if present.
    """
    if traceback:
        exception, _, message = traceback.splitlines()[-1].partition(": ")
        if exception == "OSError":
            raise OSError(message)


def data_to_writes(data: bytes, block_size: int = 512) -> Iterator[tuple[str, int]]:
    """
    Given some bytes to write to a file using calls to a function named 'w'
    in blocks of up to block_size bytes, generates a series of Python
    snippets to do the job, using hex or Python string representations
    depending on what is most efficient.

    Specifically, yields a series of (python_snippet, bytes_written)
    tuples.

    The function 'uh' is expected to be an alias to unhexlify.
    """
    while data:
        block = data[:block_size]
        data = data[block_size:]

        as_bytes = f"w({block!r})"
        len_as_bytes = len(as_bytes)
        len_as_hex = len('w(uh(b""))') + (len(block) * 2)

        # Write using whichever format is most efficient for this block
        if len_as_bytes < len_as_hex:
            yield as_bytes, len(block)
        else:
            yield f"w(uh({hexlify(block)!r}))", len(block)


def combine_sm_operations(
    operations: Iterable[tuple[str, int, int, int, int]],
    equal_overhead: int = 0,
    seek_overhead: int = 0,
) -> Iterator[tuple[str, int | None, int | None, int | None, int | None]]:
    """
    Filters/modifies a SequenceMatcher's opcodes sequence to leave only insert
    and equal operations (with all non-referenced data being implicitly
    considered deleted). Further, insert only defines j1 and j2 and equal only
    defines i1 and i2.

    Attempts to coallesce equivalent operations (e.g. insert followed by
    replace).

    The equal_overhead parameter gives the overhead of inserting an 'equal'
    operation vs just adding more values to an already ongoing 'insert'. This
    is used to decide whether to switch from an insertion to an equal.

    The seek_overhead parameter indicates additional cost to add to
    equal_overhead to account for seeking past deleted content.
    """
    # The most recent opcode to be read (if an insertion, this might be
    # extended as a result of merging adjacent insertion or equal operations)
    #
    # We begin with a 'fake' empty insertion. This enables us to handle the
    # case where the first real operation is a very short 'equal' followed by
    # an insertion without having to code up the logic for merging an insertion
    # with an equal (as opposed to the other way around which we do implement).
    cur_opcode_is_real = False
    cur_opecode = "insert"
    cur_i1: int | None = None
    cur_i2: int | None = None
    cur_j1: int | None = 0
    cur_j2: int | None = 0

    # The implicit index of the next value to read from the old input array
    # (i.e. for which no seek would be required).
    read_offset = 0

    for opcode, i1, i2, j1, j2 in operations:
        if opcode in ("insert", "replace"):
            # Replace is equiv to insert+delete and in our case we don't care
            # about deletions, therefore replace = insert for our purposes.
            opcode = "insert"

            if cur_opecode == opcode:
                assert cur_j2 == j1  # Sanity check
                cur_j2 = j2
                cur_opcode_is_real = True
            else:
                if cur_opcode_is_real:
                    yield (cur_opecode, cur_i1, cur_i2, cur_j1, cur_j2)
                cur_opcode_is_real = True
                cur_opecode = opcode
                cur_i1, cur_i2, cur_j1, cur_j2 = None, None, j1, j2
        elif opcode == "equal":
            # NB: We should never encounter two 'equal' operations in a row so
            # we'll not bother trying to merge them here.

            overhead = equal_overhead + (seek_overhead if read_offset != i1 else 0)

            if cur_opecode == "insert" and i2 - i1 <= overhead:
                # The ongoing operation is an insertion and it would be cheaper
                # to add the literal equal bytes to that than to switch to an
                # 'equal' here.
                assert cur_j2 == j1  # Sanity check
                cur_j2 = j2
                cur_opcode_is_real = True
            else:
                # Either the previous operation wasn't an insertion or this
                # equal block is large enough that we should emit it as an
                # equal directly.
                if cur_opcode_is_real:
                    yield (cur_opecode, cur_i1, cur_i2, cur_j1, cur_j2)
                cur_opcode_is_real = True
                cur_opecode = opcode
                cur_i1, cur_i2, cur_j1, cur_j2 = i1, i2, None, None
                read_offset = i2
        elif opcode == "delete":
            # Deletions are considered a no-operation. Note that we don't end
            # the previously running operation since the only operation we
            # continue are inserts and these are not sensitive to deletion.
            pass
        else:
            assert False, f"Unexpected opcode: {opcode}"

    # Flush any remaining operation
    if cur_opcode_is_real:
        yield (cur_opecode, cur_i1, cur_i2, cur_j1, cur_j2)


def data_to_update_commands(
    old_content: bytes,
    new_content: bytes,
    block_size: int = 512,
    hasher: Optional[Callable[[bytes], None]] = None,
) -> Iterator[tuple[str, int]]:
    """
    Given the existing contents of an existing  file (old_content) and
    desired new file contents (new_content), generates a series of read and
    seek operations on an old file and write operations on a new file.

    Specifically, generates a series of (python_snippet, bytes_written)
    tuples where the python_snippets should be evaluated in the given
    order and will cause the given number of bytes to be written.

    The read function on the old file is expected to be called 'r'.

    The seek function on the old file is expected to be called 's'.

    The write function on the new file expected to be called 'w'.

    The old and new files are expected to be distinct.

    Writes of new material will be restricted to at most block_size bytes
    at a time.

    If a hasher is provided, it will be called on every block of
    old_content used in the patching process.
    """
    # Guess whether we'll be using hex mode for most of the data
    sample_length_as_byte_string = len(repr(new_content[:block_size])) - 3
    sample_length_as_hex_string = len(new_content[:block_size]) * 2
    probably_mostly_hex = sample_length_as_byte_string > sample_length_as_hex_string

    # The basic overhead computed below assumes literal values are being
    # encoded as a byte string at about one byte on the wire per useful byte.
    #
    # In the hex case, however, the overhead is effectively halved because the
    # cost of encoding each byte in hex is double the length of the byte.
    overhead_scale = 2 if probably_mostly_hex else 1

    # We use the SequenceMatcher utility to produce a series of edits which
    # can convert an on-disk copy of old_content into new_content. Whilst
    # this frequently produces compact patches, certain types of edit could
    # potentially be better supported by a different algorithm, for example
    # shuffling lines in a file.
    operations = combine_sm_operations(
        SequenceMatcher(None, old_content, new_content).get_opcodes(),
        # We presume that most seeks will have the same number of digits as the
        # index half way through -- which on average they will more or less.
        seek_overhead=len(f"s({len(old_content) // 2})\n") // overhead_scale,
        # We presume that (for the sake of overhead calculations which only
        # matter for short reads) reads will be short (i.e. a single digit
        # number of bytes)
        equal_overhead=len("w(r(9))\n") // overhead_scale,
    )

    read_offset = 0
    for operation, i1, i2, j1, j2 in operations:
        if operation in ("insert", "replace"):
            assert j1 is not None
            assert j2 is not None
            # NB: In the case of replace, we'll seek past the old content
            # when we next need read.
            yield from data_to_writes(new_content[j1:j2], block_size)
        elif operation == "equal":
            assert i1 is not None
            assert i2 is not None

            if hasher is not None:
                hasher(old_content[i1:i2])

            if read_offset != i1:
                read_offset = i1
                yield (f"s({i1})", 0)
            while read_offset < i2:
                length = min(i2 - read_offset, block_size)
                yield (f"w(r({length}))", length)
                read_offset += length
        elif operation == "delete":
            # This is a no-op! We'll seek pasth the old content when we
            # next read.
            pass


def batch_commands(
    command_iter: Iterable[tuple[str, int]],
    bytes_per_batch: int = 512,
    commands_per_batch: int = 20,
) -> Iterator[tuple[str, int]]:
    """
    Group a series of commands (given as (command, bytes_touched) tuples),
    group these into batches (concatenated with newlines) where each batch
    never exceeds the given number of bytes touched or number of commands
    executed.
    """
    this_batch: list[str] = []
    this_batch_bytes = 0
    for command, num_bytes in command_iter:
        # If this command will not fit in the current batch, flush that batch
        # out
        if (
            this_batch_bytes + num_bytes > bytes_per_batch
            or len(this_batch) + 1 > commands_per_batch
        ):
            yield ("\n".join(this_batch), this_batch_bytes)
            this_batch = []
            this_batch_bytes = 0

        this_batch.append(command)
        this_batch_bytes += num_bytes

    # Output the final batch (if non-empty)
    if this_batch_bytes > 0:
        yield ("\n".join(this_batch), this_batch_bytes)


class UpdateError(Exception):
    """
    Thrown when :py:meth:`FilesystemAPI.patch` fails due to the file on disk
    differing from the one used on the host to apply the patch.
    """


class PathType(Enum):
    """Enumerates the types a path can have."""

    file = "file"
    dir = "dir"
    absent = "absent"

    def is_file(self) -> bool:
        return self == PathType.file

    def is_dir(self) -> bool:
        return self == PathType.dir


class FilesystemAPI:
    # Snippets which may be executed to define/import a useful function/module.
    #
    # {name: (script, dependencies), ...}
    _DEFINITIONS: dict[str, tuple[str, list[str]]] = {
        "os": ("import os", []),
        "time": ("import time", []),
        "sha256": ("from hashlib import sha256", []),
        # a mkdir which includes parents and exist_ok arguments
        "mkdir": (
            """
                def mkdir(path, parents, exist_ok):
                    if parents:
                        parent = ""
                        for part in path.split("/")[1:-1]:
                            parent += "/" + part
                            try:
                                os.mkdir(parent)
                            except OSError:
                                pass
                    
                    try:
                        os.mkdir(path)
                    except OSError:
                        if not exist_ok:
                            raise
            """,
            ["os"],
        ),
        # Recursively delete
        #
        # Note that this process can take a surprisingly long time for large
        # hierarchies. As such, to avoid the connection timing out, this
        # function will raise a timeout error
        "remove_recursive": (
            """
                def remove_recursive(path, timeout_ms, _timeout_at=None):
                    if _timeout_at is None:
                        _timeout_at = time.ticks_add(time.ticks_ms(), timeout_ms)
                    if os.stat(path)[0] & 0x4000:
                        for entry in os.ilistdir(path):
                            name, type = entry[:2]
                            if type & 0x4000:
                                remove_recursive(f"{path}/{name}", timeout_ms, _timeout_at)
                            else:
                                os.remove(f"{path}/{name}")
                            if time.ticks_diff(_timeout_at, time.ticks_ms()) <= 0:
                                raise Exception("Timeout")
                        os.rmdir(path)
                    else:
                        os.remove(path)
            """,
            ["os", "time"],
        ),
        # List directory and filenames within a path
        "ls": (
            """
                def ls(path):
                    directories = []
                    files = []
                    for entry in os.ilistdir(path):
                        name, type = entry[:2]
                        if type & 0x4000:
                            directories.append(name)
                        else:
                            files.append(name)
                    return (directories, files)
            """,
            ["os"],
        ),
        # Get the size of a file in bytes
        "file_len": (
            """
                def file_len(path):
                    with open(path, "rb") as f:
                        f.seek(0, 2)
                        print(f.tell())
            """,
            [],
        ),
        # Print next (approx) n-bytes worth of string literals from an iterator
        # as a Python list literal. Prints an empty list when the iterator is
        # exhausted.
        #
        # The utility of this is in printing out (possibly long) sequences of
        # of (relatively short) strings without a read timing out due to
        # the sheer time spent printing things.
        "pns": (
            """
                def pns(iterator, size):
                    so_far = 0
                    print("[", end="")
                    while so_far < size:
                        try:
                            value = next(iterator)
                            print(repr(value), end=",")
                            so_far += len(value)
                        except StopIteration:
                            break
                    print("]", end="")
            """,
            [],
        ),
        # Short aliases for unhexlify and hexlify
        "uh": ("from binascii import unhexlify as uh", []),
        "h": ("from binascii import hexlify as h", []),
        # Short alias which wraps a reader function such that all read values
        # will be added to the provided hasher.
        "make_read_and_hash": (
            """
                def make_read_and_hash(reader, hasher):
                    def wrapped_reader(*args):
                        data = reader(*args)
                        hasher.update(data)
                        return data
                    return wrapped_reader
            """,
            [],
        ),
        # Function which takes a bytes object and either returns a Python bytes
        # literal (as a string) or a Python expression calling unhexlify
        # (aliased to 'uh'), depending which is shorter.
        "bytes_to_evalable": (
            """
                def bytes_to_evalable(data):
                    as_bytes = repr(data)
                    len_as_bytes = len(as_bytes)
                    len_as_hex = len('uh(b"")') + (len(data) * 2)
                    if len_as_bytes < len_as_hex:
                        return as_bytes
                    else:
                        return f"uh({h(data)})"
            """,
            ["h"],
        ),
        # Print the next n bytes from a file using bytes_to_evalable.
        "pnb": (
            """
                def pnb(f, n):
                    print(bytes_to_evalable(f.read(n)))
            """,
            ["bytes_to_evalable"],
        ),
        # Print an unused temporary filename starting the the given prefix.
        # Does not create any files
        "get_temp_file_name": (
            """
                def get_temp_file_name(prefix):
                    i = 0
                    while True:
                        name = "{prefix}.{i}"
                        try:
                            os.stat(name)
                        except OSError:
                            print(name)
                            return
                        i += 1
            """,
            ["os"],
        ),
    }

    def __init__(self, conn: Connection) -> None:
        self._conn = conn
        self._defined: set[str] = set()

    def _ensure_defined(self, name: str) -> None:
        """
        Ensure the specified definition is in scope.
        """
        if name not in self._defined:
            self._defined.add(name)

            definition, dependencies = FilesystemAPI._DEFINITIONS[name]

            # Resolve dependencies
            for dependency in dependencies:
                self._ensure_defined(dependency)

            # Load definition
            assert raw_paste_exec(
                self._conn,
                dedent(definition).strip(),
            ) == ("", "")

    def get_type(self, path: str) -> PathType:
        """
        Determine what kind of object is at the provided path.
        """
        self._ensure_defined("os")
        out, err = raw_paste_exec(self._conn, f"print(os.stat({path!r})[0])")
        try:
            traceback_to_oserror(err)
        except OSError as exc:
            if exc.args[0] == "[Errno 2] ENOENT":
                return PathType.absent
            else:
                raise
        assert err == "", err

        mode = int(out)
        if mode & 0x4000:
            return PathType.dir
        else:
            return PathType.file

    def mkdir(
        self,
        path: str,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        """
        Create a directory.
        """
        path = path.rstrip("/")
        self._ensure_defined("mkdir")
        out, err = raw_paste_exec(self._conn, f"mkdir({path!r}, {parents}, {exist_ok})")
        traceback_to_oserror(err)
        assert err == "", err
        assert out == "", out

    def remove_recursive(self, path: str):
        """
        Remove a given file or directory tree recursively.
        """
        self._ensure_defined("remove_recursive")

        # NB: To avoid issues with remove_recursive taking so long that the
        # connection times out, the remove_recursive process will timeout
        # after a given number of ms by throwing an Exception. As such, all we
        # need to do is continue re-running the command until we nolonger get
        # that exception.
        timeout_ms = 1000
        if self._conn.timeout is not None:
            timeout_ms = int(self._conn.timeout * 500)
        while True:
            out, err = raw_paste_exec(
                self._conn, f"remove_recursive({path!r}, {timeout_ms})"
            )
            if err.endswith("\r\nException: Timeout\r\n"):
                continue
            traceback_to_oserror(err)
            assert err == "", err
            assert out == "", out
            break

    def ls(self, path: str, block_size: int = 512) -> tuple[list[str], list[str]]:
        """
        List the directories and files (separately) at a given path.
        """
        # NB: The process of reading the lists of directories and files is
        # slightly more complicated than just evaling the repr'd directory and
        # file lists because if these are long, the print command might take
        # long enough to trigger a timeout. Instead we use the 'pns' utility
        # function to print a few entries at a time.

        self._ensure_defined("ls")
        out, err = raw_paste_exec(self._conn, f"d, f = map(iter, ls({path!r}))")
        traceback_to_oserror(err)
        assert err == "", err
        assert out == "", out

        self._ensure_defined("pns")
        directories: list[str] = []
        files: list[str] = []
        for lst, name in [(directories, "d"), (files, "f")]:
            while True:
                out, err = raw_paste_exec(self._conn, f"pns({name}, {block_size})")
                assert err == "", err
                to_add = eval(out)
                if to_add:
                    lst.extend(to_add)
                else:
                    break

        return directories, files

    def rename(self, old_path: str, new_path: str) -> None:
        """
        Rename/move a file.
        """
        self._ensure_defined("os")
        out, err = raw_paste_exec(self._conn, f"os.rename({old_path!r}, {new_path!r})")
        traceback_to_oserror(err)
        assert out == "", out
        assert err == "", err

    def _open_file(
        self,
        path: str,
        mode: str,
        aliases: dict[str, str] = {},
        file_object_name: str = "f",
    ) -> None:
        """
        Open the named file, checking for OSError and assigning the open file
        object to the name ``file_object_name``.

        Also creates aliases to named members of the file object. (e.g. {'w':
        'write'} creates an alias, 'w' to 'f.write').
        """
        defs = ";".join(
            f"{alias}={file_object_name}.{member}" for alias, member in aliases.items()
        )
        out, err = raw_paste_exec(
            self._conn,
            f"{file_object_name} = open({path!r}, {mode!r});{defs}",
        )
        traceback_to_oserror(err)
        assert err == "", err
        assert out == "", out

    def _close_file(self, file_object_name: str = "f") -> None:
        """
        Close the file ``file_object_name``.
        """
        out, err = raw_paste_exec(self._conn, f"{file_object_name}.close()")
        traceback_to_oserror(err)
        assert err == "", err
        assert out == "", out

    def write_file(self, path: str, content: bytes, block_size: int = 512) -> None:
        """
        Write a file, block_size bytes at a time.

        Internally attempts to write the file using whichever mode is most
        efficient: Python bytes literals or hex strings. For mainly text-based
        files the former is typically used (with a relatively low overhead).
        For binary files, the latter will likely be used with a roughly 2x
        overhead in data transfer.
        """
        self._open_file(path, "wb", {"w": "write"})

        self._ensure_defined("uh")  # = unhexlify
        for write_block, _ in data_to_writes(content, block_size):
            assert raw_paste_exec(self._conn, write_block) == ("", "")

        self._close_file()

    def read_file(self, path: str, block_size: int = 512) -> bytes:
        """
        Read a file block_size bytes at a time.

        See :py:meth:`write_file` for the analgous process and a
        description of how this is done
        """
        self._open_file(path, "rb")

        self._ensure_defined("pnb")

        data = b""
        while True:
            out, err = raw_paste_exec(self._conn, f"pnb(f, {block_size})")
            assert err == "", err
            block = eval(out, {"uh": unhexlify})
            data += block

            if len(block) < block_size:
                break

        self._close_file()

        return data

    def file_len(self, path: str) -> int:
        """
        Get the size of a file in bytes.
        """
        self._ensure_defined("file_len")
        out, err = raw_paste_exec(self._conn, f"file_len({path!r})")
        assert err == "", err

        return int(out)

    def sync(self) -> None:
        """
        Sync the filesystem to storage, if sync is supported by the platform.
        Otherwise, does nothing.
        """
        self._ensure_defined("os")
        out, err = raw_paste_exec(self._conn, "if hasattr(os, 'sync'): os.sync()")
        assert out == "", out
        assert err == "", err

    def update_file(
        self,
        path: str,
        old_content: bytes,
        new_content: bytes,
        block_size: int = 512,
        block_transaction_limit: int = 20,
        safe: bool = True,
    ) -> None:
        """
        Attempt to update an existing file such that its contents is changed
        from ``old_content`` to ``new_content`` without transferring the whole
        file contents over the wire, just the differences.

        The ``block_size`` controls the maximum number of bytes transferred
        over the wire (or on disk) in one go.

        The ``block_transaction_limit`` limits the number of copy/write cycles
        which can occur in a single block.

        The ``safe`` parameter enables optional checks to verify that the
        reconstructed file does actually match ``new_content`` (i.e. that the
        file originally did contain ``old_content``, or rather a close-enough
        rendition). A :py:exc:`UpdateError` will be thrown if the check fails.
        If disabled, performance will be a little better but the resulting file
        may potentially be corrupt.
        """
        # Special case: Unless in safe mode, and file exists, do nothing if old
        # == new
        if not safe and old_content == new_content and self.get_type(path).is_file():
            return

        self._open_file(path, "rb", {"r": "read", "s": "seek"}, "fi")

        # We will construct the patch in a separate (temporary) file
        self._ensure_defined("get_temp_file_name")
        temp_path, err = raw_paste_exec(self._conn, f"get_temp_file_name({path!r})")
        assert err == "", err
        self._open_file(temp_path, "wb", {"w": "write"}, "fo")

        # To implement safe mode we compute a checksum of all data read from
        # the input file. If the input file has not become (too far) out of
        # sync, the same hash computed on the host should match.
        #
        # Note that we don't need to hash unread parts of the file since these
        # have on effect on the final (patched) file and so can differ without
        # consequence.
        if safe:
            self._ensure_defined("sha256")
            self._ensure_defined("make_read_and_hash")
            out, err = raw_paste_exec(
                self._conn,
                "hash = sha256(); r = make_read_and_hash(r, hash)",
            )
            assert out == "", out
            assert err == "", err

            read_hash = sha256()

        # Apply the patch
        self._ensure_defined("uh")  # = unhexlify
        for commands, _ in batch_commands(
            data_to_update_commands(
                old_content,
                new_content,
                block_size=block_size,
                hasher=read_hash.update if safe else None,
            ),
            bytes_per_batch=block_size,
            commands_per_batch=block_transaction_limit,
        ):
            out, err = raw_paste_exec(self._conn, commands)
            assert out == "", commands + "\n---\n" + out
            assert err == "", commands + "\n---\n" + err

        self._close_file("fi")
        self._close_file("fo")

        # Check integrity
        if safe:
            self._ensure_defined("h")
            out, err = raw_paste_exec(self._conn, "print(hash.digest())")
            assert err == "", err
            actual_read_hash = eval(out)

            if read_hash.digest() != actual_read_hash:
                self.remove_recursive(temp_path)
                raise UpdateError()

        # Replace the old file with the newly constructed one
        self.rename(temp_path, path)



"""
Copy files and directories.
"""

from typing import Iterator, Iterable, cast

import sys

from argparse import ArgumentParser, Namespace

from pathlib import Path

from upyt.connection import Connection
from upyt.upy_terminal import interrupt_and_enter_repl
from upyt.upy_fs import upy_filesystem, FilesystemAPI
from upyt.cli.hybrid_filesystem_api import HybridFilesystemAPI


def split_source(source: str) -> tuple[str, str]:
    """
    Given a commandline source name (e.g. '-', 'foo/bar' or ':foo/bar') split
    it into a (prefix_plus_dirname, name) pair.

    Here prefix_plus_dirname contains the ':' if any plus the base directory
    name (e.g. 'foo/' in these examples).

    Also normalises empty paths to "." and strips trailing slashes.
    """
    # You can never be sure there aren't Windows users about...
    source = source.replace("\\", "/")

    prefix = ""
    if source.startswith(":"):
        prefix = ":"
        source = source[1:]

    # Special case for empty string == cwd
    if source == "":
        source = "."

    # Normalize-out trailing slashes (special case for root dir)
    if source != "/":
        source = source.rstrip("/")

    base_dir, slash, name = source.rpartition("/")

    return (prefix + base_dir + slash, name)


class RecursionNotAllowedError(Exception):
    """Thrown when attempting to copy a directory without recursion enabled."""


def read_sources(
    hfs: HybridFilesystemAPI,
    sources: list[str],
    recursive: bool = False,
) -> Iterator[tuple[str, bytes | None]]:
    """
    Given a list of sources to copy from, return the file names and their
    contents. Directories are listed with contents of None
    """
    # Split sources into their leading components (whose names will not be used
    # in the destination) and the rest.
    #
    # [(base_dir, name), ...]
    to_visit = [split_source(source) for source in sources]

    # (Potentially) recursively iterate over the sources
    while to_visit:
        base_dir, name = to_visit.pop()

        path = base_dir + name
        if hfs.get_type(path).is_dir():
            if recursive:
                directories, files = hfs.ls(path)
                for subpath in directories + files:
                    to_visit.append((base_dir, f"{name}/{subpath}"))
                yield (name, None)
            else:
                raise RecursionNotAllowedError(base_dir + name)
        else:
            content = hfs.read_file(base_dir + name)
            yield (name, content)


def write_single_file_to_destination(
    hfs: HybridFilesystemAPI,
    file: tuple[str, bytes],
    destination: str,
) -> None:
    """
    Write a single file (as read by :py:func:`read_sources`) into the provided
    destination. If the destination is a directory, the file is copied into it.
    Otherwise, the file is copied to the given name.
    """
    filename, data = file

    if hfs.get_type(destination).is_dir():
        hfs.write_file(f"{destination}/{filename}", data)
    else:
        hfs.write_file(destination, data)


def write_multiple_files_to_existing_directory(
    hfs: HybridFilesystemAPI,
    files: Iterable[tuple[str, bytes | None]],
    destination: str,
) -> None:
    """
    Write a series of files (as read by :py:func:`read_sources`) into the
    provided destination directory.
    """
    # NB: Sorting ensures that directories are created before thier contents
    # and subdirectories
    for filename, data in sorted(files):
        if data is None:
            hfs.mkdir(f"{destination}/{filename}", exist_ok=True)
        else:
            hfs.write_file(f"{destination}/{filename}", data)


def write_single_directory_to_non_existing_destination(
    hfs: HybridFilesystemAPI,
    files: Iterable[tuple[str, bytes | None]],
    destination: str,
) -> None:
    """
    Write a series of files contained in a single directory (as read by
    :py:func:`read_sources`) into the provided non-existing destination
    directory.
    """
    for filename, data in sorted(files):
        filename = filename.partition("/")[2]
        if data is None:
            hfs.mkdir(f"{destination}/{filename}", exist_ok=True)
        else:
            hfs.write_file(f"{destination}/{filename}", data)


def cp(
    hfs: HybridFilesystemAPI,
    sources: list[str],
    destination: str,
    recursive: bool = False,
):
    """
    Copy files between the host and MicroPython device.

    Paths starting with ":" are device paths, those without are host paths.

    Copying semantics approximately follow the POSIX ``cp`` command.
    """
    files = read_sources(hfs, sources, recursive)
    if len(sources) == 1 and not hfs.get_type(sources[0]).is_dir():
        write_single_file_to_destination(
            hfs,
            cast(tuple[str, bytes], next(files)),  # Is directory per test above
            destination,
        )
    elif len(sources) == 1 and not hfs.get_type(destination).is_dir():
        # NB: source is dir, as per the previous branch's test
        write_single_directory_to_non_existing_destination(hfs, files, destination)
    else:
        write_multiple_files_to_existing_directory(hfs, files, destination)


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "source",
        nargs="+",
        help="""
            Files to copy. Prefix with ':' for device paths.
        """,
    )
    parser.add_argument(
        "destination",
        help="""
            Destination. Prefix with ':' for device paths.
        """,
    )
    parser.add_argument(
        "--recursive",
        "-r",
        "-R",
        action="store_true",
        default=False,
        help="""
            Copy directories and their contents recursively.
        """,
    )


def main(args: Namespace):
    with Connection.from_specification(args.device) as conn:
        interrupt_and_enter_repl(conn)
        with upy_filesystem(conn) as fs:
            if not (
                any(source.startswith(":") for source in args.source)
                or args.destination.startswith(":")
            ):
                print(
                    "warning: neither source nor destination on device (i.e. starting with ':')",
                    file=sys.stderr,
                )
            hfs = HybridFilesystemAPI(fs)
            cp(hfs, args.source, args.destination, args.recursive)
            hfs.sync()



"""
Create directories on a MicroPython device.
"""

import sys

from argparse import ArgumentParser, Namespace

from upyt.connection import Connection
from upyt.upy_terminal import interrupt_and_enter_repl
from upyt.upy_fs import upy_filesystem
from upyt.cli.hybrid_filesystem_api import HybridFilesystemAPI


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "path",
        nargs="+",
        help="""
            The directories to create. Prefix with ':' for device paths.
        """,
    )

    parser.add_argument(
        "--parents",
        "-p",
        action="store_true",
        help="""
            If given, create parent directories too and don't throw an error if
            the directory already exists.
        """,
    )


def main(args: Namespace):
    with Connection.from_specification(args.device) as conn:
        interrupt_and_enter_repl(conn)
        with upy_filesystem(conn) as fs:
            if not any(path.startswith(":") for path in args.path):
                print(
                    "warning: path not on device (i.e. starting with ':')",
                    file=sys.stderr,
                )
            hfs = HybridFilesystemAPI(fs)
            for path in args.path:
                hfs.mkdir(
                    path,
                    parents=args.parents,
                    exist_ok=args.parents,
                )
            hfs.sync()



import pytest

from unittest.mock import Mock

import random

from difflib import SequenceMatcher

from upyt.upy_fs import (
    upy_filesystem,
    data_to_writes,
    combine_sm_operations,
    data_to_update_commands,
    batch_commands,
    UpdateError,
    FilesystemAPI,
    PathType,
)


class TestDataToWrites:
    def test_empty(self) -> None:
        assert list(data_to_writes(b"", block_size=2)) == []

    def test_less_than_a_block(self) -> None:
        assert list(data_to_writes(b"1234", block_size=10)) == [("w(b'1234')", 4)]
        assert list(data_to_writes(b"1234", block_size=4)) == [("w(b'1234')", 4)]

    def test_multiple_blocks(self) -> None:
        assert list(data_to_writes(b"1234", block_size=2)) == [
            ("w(b'12')", 2),
            ("w(b'34')", 2),
        ]
        assert list(data_to_writes(b"1234", block_size=3)) == [
            ("w(b'123')", 3),
            ("w(b'4')", 1),
        ]

    def test_use_hex_if_smaller(self) -> None:
        assert list(data_to_writes(b"\xFF\xFF\xFF\xFF", block_size=4)) == [
            ("w(uh(b'ffffffff'))", 4),
        ]

    def test_decide_block_by_block(self) -> None:
        assert list(data_to_writes(b"hi\xFF\xFF", block_size=2)) == [
            ("w(b'hi')", 2),
            ("w(uh(b'ffff'))", 2),
        ]


class TestCombineSMOperations:
    def test_empty(self) -> None:
        assert list(combine_sm_operations([])) == []

    def test_single_insert_operation(self) -> None:
        assert list(
            combine_sm_operations(
                SequenceMatcher(None, "", "insert-me").get_opcodes(),
            )
        ) == [
            ("insert", None, None, 0, 9),
        ]

    def test_single_replace_operation(self) -> None:
        assert list(
            combine_sm_operations(
                SequenceMatcher(None, "REPLACE_ME", "replace-me").get_opcodes(),
            )
        ) == [
            ("insert", None, None, 0, 10),
        ]

    def test_single_delete_operation(self) -> None:
        assert (
            list(
                combine_sm_operations(
                    SequenceMatcher(None, "delete-me", "").get_opcodes(),
                )
            )
            == []
        )

    def test_single_equal_operation(self) -> None:
        assert list(
            combine_sm_operations(
                SequenceMatcher(None, "equal-me", "equal-me").get_opcodes(),
            )
        ) == [
            ("equal", 0, 8, None, None),
        ]

    def test_single_equal_operation_turned_into_insert_if_short(self) -> None:
        assert list(
            combine_sm_operations(
                SequenceMatcher(None, "e", "e").get_opcodes(),
                equal_overhead=10,
            )
        ) == [
            ("insert", None, None, 0, 1),
        ]

    def test_merge_insert_and_replace_operations(self) -> None:
        # NB: Insert and replace operations are never emitted sequentially by
        # the current version of sequence matcher but this is not a guarantee
        # so we explicitly test this here.
        assert list(
            combine_sm_operations(
                [
                    ("insert", 0, 0, 0, 5),
                    ("replace", 0, 3, 5, 8),
                    ("equal", 3, 5, 8, 10),
                ]
            )
        ) == [
            ("insert", None, None, 0, 8),
            ("equal", 3, 5, None, None),
        ]

    def test_merge_short_equal_with_previous_insert(self) -> None:
        # NB: These operations are never emitted by the current version of
        # sequence matcher but this is not a guarantee so we explicitly test
        # this here.
        assert list(
            combine_sm_operations(
                [
                    ("insert", 0, 0, 0, 5),
                    # Shorter than overhead, should be merged
                    ("equal", 0, 4, 5, 9),
                    # Multiple short equals should be merged
                    ("equal", 4, 5, 9, 10),
                    # Even if interspersed with deletions -- though in such
                    # cases we'll inevetably need a seek so should be allowed
                    # to go for longer
                    ("delete", 5, 9, 10, 10),
                    ("equal", 9, 16, 10, 17),
                    # NB: we should be able to do this twice since the previous
                    # euqal should be omitted and therefore a seek should still
                    # be needed!
                    ("replace", 16, 17, 17, 18),
                    ("equal", 17, 24, 18, 25),
                    # Longer than overhead, should be inserted as an equal
                    ("equal", 24, 34, 25, 35),
                    # Check that when no seek is needed, the threshold is lower
                    # (and so the equal below should be kept as an equal)
                    ("insert", 34, 34, 35, 40),
                    ("equal", 34, 39, 40, 45),
                ],
                equal_overhead=4,
                seek_overhead=3,
            )
        ) == [
            ("insert", None, None, 0, 25),
            ("equal", 24, 34, None, None),
            ("insert", None, None, 35, 40),
            ("equal", 34, 39, None, None),
        ]

    def test_merge_initial_short_equal_into_following_insert(self) -> None:
        assert list(
            combine_sm_operations(
                [
                    ("equal", 0, 1, 0, 1),
                    ("insert", 1, 1, 1, 3),
                ],
                equal_overhead=2,
            )
        ) == [
            ("insert", None, None, 0, 3),
        ]

    def test_do_not_merge_initial_long_equal_into_following_insert(self) -> None:
        assert list(
            combine_sm_operations(
                [
                    ("equal", 0, 3, 0, 3),
                    ("insert", 3, 3, 3, 4),
                ],
                equal_overhead=2,
            )
        ) == [
            ("equal", 0, 3, None, None),
            ("insert", None, None, 3, 4),
        ]

    @pytest.mark.parametrize(
        "before, after, equal_overhead, exp_operations",
        [
            # Trivial cases
            ("", "", 0, 0),
            ("already-equal", "already-equal", 0, 1),
            ("", "all-insertion", 0, 1),
            ("all-deletion", "", 0, 0),
            ("some-REPLACEMENT", "some-replacement", 0, 2),
            # Equality merged with insertions
            ("aa-bb-cc", "AA-BB-CC", 0, 5),  # No overhead so will use insert and equal
            ("aa-bb-cc", "AA-BB-CC", 2, 1),  # Should turn into a big insertion
            # Deletions between merged-in equal blocks
            (
                "WHAT-was-HERE",
                "what--here",
                0,
                4,
            ),  # No overhead so will use insert and equal
            ("WHAT-was-HERE", "what--here", 2, 1),  # Should turn into big insertion
        ],
    )
    def test_correctness(
        self,
        before: str,
        after: str,
        equal_overhead: int,
        exp_operations: int,
    ) -> None:
        out = []
        for opcode, i1, i2, j1, j2 in combine_sm_operations(
            SequenceMatcher(None, before, after).get_opcodes(),
            equal_overhead=equal_overhead,
        ):
            if opcode == "insert":
                out.append(after[j1:j2])
                assert i1 is None
                assert i2 is None
            elif opcode == "equal":
                out.append(before[i1:i2])
                assert j1 is None
                assert j2 is None
            else:
                assert False, opcode

        # Check the substitutions actually do what they're supposed to
        assert "".join(out) == after
        assert len(out) == exp_operations


class TestDataToUpdateCommands:
    def test_empty(self) -> None:
        assert list(data_to_update_commands(b"", b"")) == []

    def test_empty_afterwards(self) -> None:
        assert list(data_to_update_commands(b"1234", b"")) == []

    def test_equal_blocking(self) -> None:
        assert list(
            data_to_update_commands(b"12345678910", b"12345678910", block_size=20)
        ) == [
            ("w(r(11))", 11),
        ]
        assert list(
            data_to_update_commands(b"12345678910", b"12345678910", block_size=11)
        ) == [
            ("w(r(11))", 11),
        ]
        assert list(
            data_to_update_commands(b"12345678910", b"12345678910", block_size=8)
        ) == [
            ("w(r(8))", 8),
            ("w(r(3))", 3),
        ]

    def test_insertion_blocking(self) -> None:
        assert list(
            data_to_update_commands(b"12345678910", b"XXX12345678910", block_size=2)
        ) == [
            ("w(b'XX')", 2),
            ("w(b'X')", 1),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(1))", 1),
        ]

    def test_seeking_after_deletion(self) -> None:
        assert list(data_to_update_commands(b"123456789101112", b"23456789101112")) == [
            ("s(1)", 0),
            ("w(r(14))", 14),
        ]

    def test_seeking_after_replacement(self) -> None:
        # NB: Note that the extra here has to be quite long to force the seek
        # to be worth it!
        assert list(
            data_to_update_commands(b"1234567891011121314", b"XX34567891011121314")
        ) == [
            ("w(b'XX')", 2),
            ("s(2)", 0),
            ("w(r(17))", 17),
        ]

    def test_no_seeking_after_insertion(self) -> None:
        assert list(
            data_to_update_commands(
                b"12345678910111213141516",
                b"XXX12345678910YYY111213141516",
            )
        ) == [
            ("w(b'XXX')", 3),
            ("w(r(11))", 11),
            ("w(b'YYY')", 3),
            ("w(r(12))", 12),
        ]

    def test_hasher(self) -> None:
        hashed: list[bytes] = []

        assert list(
            data_to_update_commands(
                b"1234567891011121314151617181920212223",
                b"XXX34567891011121314Y1617181920212223",
                block_size=2,
                hasher=hashed.append,
            )
        ) == [
            ("w(b'XX')", 2),
            ("w(b'X')", 1),
            ("s(2)", 0),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(1))", 1),
            ("w(b'Y')", 1),
            ("s(21)", 0),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
            ("w(r(2))", 2),
        ]

        # Should only hash input components which are used.
        assert hashed == [
            # NB: Implementation detial which could validly be done
            # differently: always hashes full region at once even when written
            # out in blocks
            b"34567891011121314",
            b"1617181920212223",
        ]

    @pytest.mark.parametrize(
        "old, new, exp_seek_overhead, exp_equal_overhead",
        [
            # NB: The 'old' value is what determines the seek length
            (b"X", b"Y", len("s(1)\n"), len("w(r(9))\n")),
            (b"X" * 1024, b"Y", len("s(512)\n"), len("w(r(9))\n")),
            # NB: The 'new' value is what determines whether we should assume
            # hex or not.
            (b"X" * 1024, b"not-hex", len("s(512)\n"), len("w(r(9))\n")),
            (b"X" * 1024, b"\xFF" * 10, len("s(512)\n") // 2, len("w(r(9))\n") // 2),
        ],
    )
    def test_overhead_values(
        self,
        monkeypatch,
        old: bytes,
        new: bytes,
        exp_seek_overhead: int,
        exp_equal_overhead: int,
    ) -> None:
        from upyt import upy_fs

        mock_combine_sm_operations = Mock(side_effect=combine_sm_operations)
        monkeypatch.setattr(upy_fs, "combine_sm_operations", mock_combine_sm_operations)

        list(data_to_update_commands(old, new))

        assert len(mock_combine_sm_operations.mock_calls) == 1
        call = mock_combine_sm_operations.mock_calls[0]

        assert call[2]["seek_overhead"] == exp_seek_overhead
        assert call[2]["equal_overhead"] == exp_equal_overhead


class TestBatchCommands:
    def test_empty(self) -> None:
        assert list(batch_commands([])) == []

    def test_fit_in_single_batch(self) -> None:
        assert list(batch_commands([("a", 1), ("b", 10)])) == [("a\nb", 11)]

        assert list(
            batch_commands(
                [("a", 1), ("b", 10)],
                bytes_per_batch=11,
                commands_per_batch=2,
            )
        ) == [("a\nb", 11)]

    def test_split_on_bytes(self) -> None:
        assert list(
            batch_commands(
                [("a", 51), ("b", 49)],
                bytes_per_batch=99,
            )
        ) == [("a", 51), ("b", 49)]

    def test_split_on_commands(self) -> None:
        assert list(
            batch_commands(
                [("a", 10), ("b", 20), ("c", 30)],
                commands_per_batch=2,
            )
        ) == [
            ("a\nb", 30),
            ("c", 30),
        ]


class TestPrimitives:
    def test_create_delete_empty_directory(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            # Check created
            dirs, files = fs.ls("/")
            assert tmpdir[1:] in dirs

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_mkdir_features(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            # Exists OK
            with pytest.raises(OSError):
                fs.mkdir(tmpdir, exist_ok=False)
            fs.mkdir(tmpdir, exist_ok=True)

            # Parents
            # NB: Parents existing (or not) is OK
            with pytest.raises(OSError):
                fs.mkdir(tmpdir + "/a/bc/defg", parents=False)
            fs.mkdir(tmpdir + "/a/bc/defg", parents=True)

            # Parents doesn't allow existance of the final directory
            with pytest.raises(OSError):
                fs.mkdir(tmpdir + "/a/bc/defg", parents=True, exist_ok=False)
            fs.mkdir(tmpdir + "/a/bc/defg", parents=True, exist_ok=True)

            # Check created
            assert fs.ls(tmpdir) == (["a"], [])
            assert fs.ls(f"{tmpdir}/a") == (["bc"], [])
            assert fs.ls(f"{tmpdir}/a/bc") == (["defg"], [])
            assert fs.ls(f"{tmpdir}/a/bc/defg") == ([], [])

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_read_and_write_file(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            # Bytes literal mode
            fs.write_file(f"{tmpdir}/foo", b"Hello, world!\n")
            assert fs.read_file(f"{tmpdir}/foo") == b"Hello, world!\n"

            # Hex mode
            fs.write_file(f"{tmpdir}/foo", b"\xFF" * 10)
            assert fs.read_file(f"{tmpdir}/foo") == b"\xFF" * 10

            # Chop into blocks (multiple of size)
            fs.write_file(f"{tmpdir}/foo", b"Hello, world!\n", block_size=2)
            assert fs.read_file(f"{tmpdir}/foo", block_size=2) == b"Hello, world!\n"

            # Chop into blocks (not a multiple of size)
            fs.write_file(f"{tmpdir}/foo", b"Hello, world!\n", block_size=3)
            assert fs.read_file(f"{tmpdir}/foo", block_size=3) == b"Hello, world!\n"

            # Run on something large with default block size
            fs.write_file(f"{tmpdir}/foo", b"X" * 1024 * 4)
            assert fs.read_file(f"{tmpdir}/foo") == b"X" * 1024 * 4

            # Run on something large in hex mode with default block size
            fs.write_file(f"{tmpdir}/foo", b"\xFF" * 1024 * 2)
            assert fs.read_file(f"{tmpdir}/foo") == b"\xFF" * 1024 * 2

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    @pytest.mark.parametrize("data", [b"", b"Foo"])
    def test_file_len(self, ser, data: bytes) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            fs.write_file(f"{tmpdir}/foo", data)
            assert fs.file_len(f"{tmpdir}/foo") == len(data)

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_rename_files(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            fs.write_file(f"{tmpdir}/foo", b"Hello, world!\n")

            fs.rename(f"{tmpdir}/foo", f"{tmpdir}/bar")
            assert fs.ls(tmpdir)[1] == ["bar"]
            assert fs.read_file(f"{tmpdir}/bar") == b"Hello, world!\n"

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_rename_directories(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)
            fs.mkdir(f"{tmpdir}/foo")
            fs.write_file(f"{tmpdir}/foo/a", b"I'm a")
            fs.write_file(f"{tmpdir}/foo/b", b"I'm b")

            fs.rename(f"{tmpdir}/foo", f"{tmpdir}/bar")
            assert fs.ls(tmpdir)[0] == ["bar"]
            assert set(fs.ls(f"{tmpdir}/bar")[1]) == {"a", "b"}
            assert fs.read_file(f"{tmpdir}/bar/a") == b"I'm a"
            assert fs.read_file(f"{tmpdir}/bar/b") == b"I'm b"

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_rename_bad_source(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            with pytest.raises(OSError):
                fs.rename(f"{tmpdir}/foo", f"{tmpdir}/bar")

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_rename_bad_destination(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            fs.write_file(f"{tmpdir}/foo", b"Hello, world!\n")

            with pytest.raises(OSError):
                fs.rename(f"{tmpdir}/foo", f"{tmpdir}/bar/baz")

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_ls_and_remove_recursive(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            # Make a large number of files and directories (to make sure we
            # don't encounter any timeout related issues with such large
            # numbers of files
            exp_dirs = {
                f"subdir_with_a_really_quite_long_name_you_know{i}" for i in range(15)
            }
            exp_files = {
                f"file_with_a_really_quite_long_name_you_know{i}" for i in range(15)
            }
            for dir in exp_dirs:
                fs.mkdir(f"{tmpdir}/{dir}")
            for file in exp_files:
                fs.write_file(f"{tmpdir}/{file}", b"")

            dirs, files = fs.ls(tmpdir)

            assert set(dirs) == exp_dirs
            assert set(files) == exp_files

            # Now remove (without timing out!)
            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_sync(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            # Pretty much all we can do is check this doesn't crash...
            fs.sync()

    def test_get_type(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            fs.write_file(f"{tmpdir}/a_file", b"hello")
            fs.mkdir(f"{tmpdir}/a_dir")

            assert fs.get_type(f"{tmpdir}/a_file") == PathType.file
            assert fs.get_type(f"{tmpdir}/a_dir") == PathType.dir
            assert fs.get_type(f"{tmpdir}/absent") == PathType.absent
            assert fs.get_type(f"{tmpdir}/deep/absent") == PathType.absent

            fs.remove_recursive(tmpdir)

    def test_update_file_basic_case(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            old_content = b"Hello there!"
            new_content = b"Hello, world!"
            fs.write_file(f"{tmpdir}/foo", old_content)
            fs.update_file(f"{tmpdir}/foo", old_content, new_content)
            assert fs.read_file(f"{tmpdir}/foo") == new_content

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_update_file_large_edit(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            old_content = b"\xFF"
            new_content = b"\xFF" * 4 * 1024
            fs.write_file(f"{tmpdir}/foo", old_content)
            fs.update_file(f"{tmpdir}/foo", old_content, new_content)
            assert fs.read_file(f"{tmpdir}/foo") == new_content

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_update_file_many_edits(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            old_content = b"\xFF" * 4 * 1024
            new_content = b"".join(
                bytes([c]) if (i % 2) else b"\00" for i, c in enumerate(old_content)
            )
            fs.write_file(f"{tmpdir}/foo", old_content)
            fs.update_file(f"{tmpdir}/foo", old_content, new_content)
            assert fs.read_file(f"{tmpdir}/foo") == new_content

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_update_safe_mode_off(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            fs.write_file(f"{tmpdir}/foo", b"HELLO?")
            # Should get a mangled file but shouldn't crash when safe-mode
            # disabled.
            fs.update_file(
                f"{tmpdir}/foo",
                b"Hello there how are you?",
                b"Hello there how are we?!",
                safe=False,
            )
            assert fs.read_file(f"{tmpdir}/foo") != b"HELLO?"

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs

    def test_update_wrong_old_content(self, ser) -> None:
        with upy_filesystem(ser) as fs:
            tmpdir = f"/d{random.randint(0, 10000)}"
            fs.mkdir(tmpdir)

            fs.write_file(f"{tmpdir}/foo", b"NOPE!")
            with pytest.raises(UpdateError):
                fs.update_file(
                    f"{tmpdir}/foo",
                    b"Hello there how are you?",
                    b"Hello there how are we?!",
                    safe=True,
                )

            # Shouldn't have changed the file on failure
            assert fs.read_file(f"{tmpdir}/foo") == b"NOPE!"

            # Shouldn't have left any temporary files behind
            assert fs.ls(tmpdir) == ([], ["foo"])

            fs.remove_recursive(tmpdir)

            # Check deleted
            dirs, files = fs.ls("/")
            assert tmpdir[1:] not in dirs



"""
Efficiently synchronise a local directory to a MicroPython device. With
'--terminal', provides an integrated serial terminal where 'Ctrl+R' re-runs
synchronisation.
"""

from argparse import ArgumentParser, Namespace

import sys

from pathlib import Path
from functools import partial

from upyt.connection import Connection
from upyt.upy_repl import expect, interrupt_and_enter_repl
from upyt.upy_terminal import GREY, RESET
from upyt.upy_fs import upy_filesystem
from upyt.sync import sync_to_device, default_exclude

from upyt.cli import terminal


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "source",
        help="""
            Local directory to synchronise to the device.
        """,
    )
    parser.add_argument(
        "destination",
        nargs="?",
        default=":/",
        help="""
            Location of corresponding directory on device. Must start with ':'.
            Defaults to the root of the device (i.e. ':/'). Note, existing
            files will be updated but files which are deleted on the host will
            *not* be automatically deleted on the device.
        """,
    )
    parser.add_argument(
        "--exclude",
        action="append",
        default=[],
        help="""
            A rsync-style exclusion pattern for paths to be excluded from sync.
            May be used multiple times. Unless --no-default-exclusions is
            given, common temporary and version control files and directories
            are excluded by default.
        """,
    )
    parser.add_argument(
        "--no-default-exclusions",
        "-E",
        action="store_true",
        default=False,
        help="""
            If given, don't exclude any temporary or version control files or
            directories by default.
        """,
    )
    parser.add_argument(
        "--force-enumerate-files",
        "-f",
        action="store_true",
        default=False,
        help="""
            If given, always scans the device to check for missing files.
        """,
    )
    parser.add_argument(
        "--force-safe-update",
        "--safe",
        "-s",
        action="store_true",
        default=False,
        help="""
            If given, always verify that modified files on the device have been
            edited correctly. Only necessary if files may have been changed on
            the device.
        """,
    )
    parser.add_argument(
        "--reset",
        "-r",
        action="store_true",
        default=False,
        help="""
            If given, reset the device after syncing (e.g. so modules are
            reloaded and new code runs). Otherwise, the device will be left at
            the REPL following file sync.
        """,
    )
    parser.add_argument(
        "--terminal",
        "-t",
        action="store_true",
        default=False,
        help="""
            If given, start the serial terminal after syncing. Synchronisation
            can be re-run at any time by pressing ctrl+r.
        """,
    )
    terminal.add_arguments(parser)


def print_progress(
    current_file: Path,
    files_to_update: set[Path],
    all_files: set[Path],
    terminal_mode: bool,
) -> None:
    if terminal_mode:
        prefix = f"{GREY}    "
        suffix = RESET
    else:
        prefix = suffix = ""

    print(f"{prefix}{current_file}...{suffix}")


def main(args: Namespace):
    # Check source/destination are local/device respectively
    if args.source.startswith(":"):
        print(
            "error: source path must be on host (i.e. not start with ':')",
            file=sys.stderr,
        )
        sys.exit(1)
    if not args.destination.startswith(":"):
        print(
            "error: destination path must be on device (i.e. start with ':')",
            file=sys.stderr,
        )
        sys.exit(1)

    # Compile exclusions list
    if args.no_default_exclusions:
        exclude = []
    else:
        exclude = default_exclude
    exclude += args.exclude

    # Select progress display
    if args.quiet:
        progress_callback = None
    else:
        progress_callback = partial(print_progress, terminal_mode=args.terminal)

    with Connection.from_specification(args.device) as conn:
        first_run = True
        while True:
            if args.terminal and not args.quiet:
                print(f"{GREY}Synchronising files:{RESET}")

            interrupt_and_enter_repl(conn)
            with upy_filesystem(conn) as fs:
                sync_to_device(
                    fs,
                    host_dir=Path(args.source),
                    device_dir=args.destination[1:],  # Strip off ':'
                    exclude=exclude,
                    force_enumerate_files=args.force_enumerate_files,
                    force_safe_update=args.force_safe_update,
                    progress_callback=progress_callback,
                )
                fs.sync()
                if not args.terminal:
                    return

            if not args.quiet:
                print(f"{GREY}Done{RESET}")

            if args.reset:
                conn.write(b"\x04")  # ctrl+d (trigger reset)
            else:
                conn.write(b"\r\n")  # Force prompt to be shown
            expect(conn, b"\r\n")

            exit_seq = terminal.terminal(
                conn,
                args,
                extra_exit_on=["\x12"],  # ctlr+r"
                extra_help="Press Ctrl+R to re-run file sync.",
                show_help=first_run,
            )
            if exit_seq != "\x12":  # != ctrl+r
                return
            first_run = False

            interrupt_message = interrupt_and_enter_repl(conn)
            if interrupt_message:
                sys.stdout.buffer.write(interrupt_message + b"\n")



"""
An implementation of a serial terminal for micropython devices.

**Currently this implementation is Unix-like only...**
"""

from typing import TextIO, Callable, Iterator, Any

import os
import sys
import selectors
import copy
import termios
import codecs

from contextlib import contextmanager, nullcontext

from upyt.connection import Connection
from upyt.read_proxy import ReadProxy, match, StreamFilter
from upyt.upy_repl import (
    expect,
    interrupt_and_enter_repl,
    paste_exec,
    MicroPythonReplError,
)

# ANSI escape sequences used internally
GREY = "\033[90m"
RESET = "\033[0m"

CLEAR = "\033[2J"
CURSOR_HOME = "\033[H"

BRACKETED_PASTE_ENABLE = "\033[?2004h"
BRACKETED_PASTE_DISABLE = "\033[?2004l"
BRACKETED_PASTE_BEGIN = "\033[200~"
BRACKETED_PASTE_END = "\033[201~"


@contextmanager
def terminal_mode(terminal: TextIO = sys.stdin):
    """
    For the duration of the context manager, disables echo, line buffering and
    conversion of keyboard interrupts into signals. Restores termios state on
    exit.
    """
    old = termios.tcgetattr(terminal)
    new = copy.deepcopy(old)

    new[3] = (
        new[3]
        # Disable canonical (e.g. the common line-buffered mode)
        & ~termios.ICANON
        # Disable local echo
        & ~termios.ECHO
        # Disable translation of interrupt keyboard sequences (e.g. Ctrl+C)
        # into signals
        & ~termios.ISIG
    )

    # Set the minimum read size to 1 byte
    new[6][termios.VMIN] = 1

    # Disable timeout while waiting for a byte
    new[6][termios.VTIME] = 0

    termios.tcsetattr(terminal, termios.TCSANOW, new)
    try:
        yield
    finally:
        termios.tcsetattr(terminal, termios.TCSANOW, old)


@contextmanager
def bracketed_paste_mode(stdout: TextIO = sys.stdout) -> Iterator[None]:
    """
    Context manager which enables bracketed paste mode in the terminal on entry
    and disables it again on exit.
    """
    stdout.write(BRACKETED_PASTE_ENABLE)
    stdout.flush()
    try:
        yield
    finally:
        stdout.write(BRACKETED_PASTE_DISABLE)
        stdout.flush()


@contextmanager
def nonblocking_file(stream: TextIO = sys.stdin):
    """
    Configures a given stream in non-blocking mode for the duration of the
    context manager, restoring the previous state on exit.
    """
    old = os.get_blocking(stream.fileno())
    os.set_blocking(stream.fileno(), False)
    try:
        yield
    finally:
        os.set_blocking(stream.fileno(), old)


@contextmanager
def nonblocking_connection(conn: Connection):
    """
    Configures a given serial connection in non-blocking mode for the duration
    of the context manager, restoring the previous state on exit.
    """
    old = conn.timeout
    conn.timeout = 0
    try:
        yield
    finally:
        conn.timeout = old


class ExitTerminal(Exception):
    """Used internally within :py:class:`raw_serial_terminal`."""

    @classmethod
    def make_raise_cb(cls, match_string: str) -> Callable[[], None]:
        def callback() -> None:
            raise cls(match_string)

        return callback


def raw_serial_terminal(
    conn: Connection,
    stdin: TextIO = sys.stdin,
    stdout: TextIO = sys.stdout,
    exit_on: list[str] = ["\x1d"],  # Default = Ctrl+]
) -> str:
    """
    A bare-minimum implementation of a MicroPython compatible serial terminal.

    Things this does:

    * Setup terminal in non-echo mode
    * Translate system text encoding to/from UTF-8 for the MicroPython device
    * Translate line-endings to/from CRLF for the MicroPython device
    * Setup terminal to capture Ctrl+C as a keyboard input (so that it is
      forwarded to MicroPython rather than triggering a KeyboardInterrupt
      here).
    * Restore terminal configuration on exit or if an exception is thrown.

    This function will gracefully exit if any of the strings in the 'exit_on'
    list are found in the stdin stream (i.e. are typed by the user). The
    matched string is returned by this function and *not* forwarded to the
    MicroPython device.
    """
    # Wraps the stdin stream such that when any of the exit strings are
    # encountered in the stream, an ExitTerminal exception is thrown containing
    # the discovered match.
    stdin = ReadProxy(stdin, [match(s, ExitTerminal.make_raise_cb(s)) for s in exit_on])

    try:
        with (
            terminal_mode(stdin),
            nonblocking_file(stdin),
            nonblocking_connection(conn),
            selectors.DefaultSelector() as sel,
        ):
            sel.register(conn, selectors.EVENT_READ)
            sel.register(stdin, selectors.EVENT_READ)

            # The MicroPython shell expects/produces UTF-8 (well actually it
            # expects ASCII, for now, but may one day accept UTF-8). We'll use an
            # incremental decoder to parse the incoming stream since we may not
            # receieve whole code points in every message.
            utf_8_decoder = codecs.lookup("utf-8").incrementaldecoder(errors="replace")

            while True:
                for key, _events in sel.select():
                    if key.fileobj is stdin:
                        c = stdin.read()
                        c = c.replace("\n", "\r\n")  # Use DOS-style newlines
                        conn.write(c.encode("utf-8"))
                        conn.flush()
                    elif key.fileobj is conn:
                        c = utf_8_decoder.decode(conn.read_buffered())
                        c = c.replace("\r", "")  # Restore UNIX-style newlines
                        stdout.write(c)
                        stdout.flush()
    except ExitTerminal as exc:
        return exc.args[0]


def handle_bracketed_paste(
    conn: Connection,
    stdin: TextIO = sys.stdin,
    stdout: TextIO = sys.stdout,
) -> None:
    """
    Assuming a BRACKETED_PASTE_BEGIN sentinel has just been read from stdin,
    read the rest of the pasted value and then send it to MicroPython.

    For single-line pastes, simply sends the string verbatim to MicroPython.
    Upon resuming the serial terminal process, the pasted text should be echoed
    back by the MicroPython REPL.

    For multi-line pastes, uses paste mode (interrupting any ongoing line
    editing operation or running code). A dummy representation of the
    paste-mode echoed text is printed to stdout whilst the actual paste-mode
    echo response from the REPL is captured for flow-control purposes. The
    first bytes to be read from MicroPython after this function returns will be
    the output of the pasted code.
    """
    # Read remainder of bracketed paste
    buffer = ""
    while not buffer.endswith(BRACKETED_PASTE_END):
        buffer += stdin.read(1)
    content = buffer[: -len(BRACKETED_PASTE_END)]

    # NB: We consider 'single line' to mean anything where all the
    # non-whitespace occurs on a single line -- leading/trailing
    # whitespace are ignored.
    single_line = len(content.strip("\r\n").splitlines()) == 1

    if single_line:  # Paste as-is
        conn.write(content.replace("\n", "\r\n").encode("utf-8"))
    else:  # Use paste mode
        try:
            # Ensure we have a clean prompt (e.g. no half-finished line edits)
            interruption_output = interrupt_and_enter_repl(conn)
            stdout.write(interruption_output.decode("utf-8", "replace"))
            stdout.flush()

            # Paste (and print mock output to terminal)
            paste_exec(conn, content)
            stdout.write("\n")
            stdout.write("".join(f"=== {line}\n" for line in content.splitlines()))
            stdout.flush()
        except MicroPythonReplError as e:
            print(
                f"\n{GREY}upyt error: Failed paste using paste mode ({e!r}).{RESET}",
                file=stdout,
                flush=True,
            )


def handle_ctrl_l_emulation(
    conn: Connection, stdout: TextIO, timeout: float = 0.1
) -> None:
    """
    Crudely emulate the typical clear-screen behaviour of a terminal by
    clearing the display and pressing return to get a new prompt.
    """
    # Clear the screen
    stdout.write(f"{CLEAR}{CURSOR_HOME}")
    stdout.flush()

    # Press enter to get a new REPL
    conn.write(b"\r\n")

    # Absorb the echoed newline to prevent the new prompt being printed on the
    # second line of the newly cleared screen.
    old_timeout = conn.timeout
    try:
        # We might not be in the REPL so the newline is not guaranteed to get
        # echoed back to us, hence the short timeout here.
        conn.timeout = timeout
        expect(conn, b"\r\n")
    except MicroPythonReplError:
        pass
    finally:
        conn.timeout = old_timeout


def serial_terminal(
    conn: Connection,
    stdin: TextIO = sys.stdin,
    stdout: TextIO = sys.stdout,
    exit_on: list[str] = ["\x1d"],  # Default = Ctrl+]
    automatic_paste_mode: bool = True,
    emulate_ctrl_l: bool = True,
) -> str:
    """
    Implements a serial terminal for the MicroPython REPL, with a handful of
    optional niceties to make it feel more like interacting with a normal
    Python shell.

    The terminal will run until a keyboard sequence listed in the `exit_on`
    argument is encountered. (By default this is Ctrl+]). This function will
    return the matched key sequence which caused it to exit.

    Parameters
    ==========
    conn : Connection
    stdin : TextIO
    stdout : TextIO
        The connection to the MicroPython device and (text mode) open
        stdin/stdout streams attached to the virtual terminal.
    exit_on : ["text", ...]
        A list of text sequences received from stdin to cause the terminal to
        exit. The matched sequence will not be sent to the device but will be
        returned by this function.
    automatic_paste_mode : bool
        If enabled, automatically write multi-line strings pasted into the
        terminal using the MicroPython REPL's paste mode.
    emulate_ctrl_l : bool
        If enabled, not-quite-emulates typical clear-screen-on-Ctrl+L
        behaviour.  Causes Ctrl+L to clear the terminal, moving the cursor to
        the top left, sending a 'return' keypress to cause a new prompt to be
        printed.
    """
    paste_mode_context_manager: Any
    if automatic_paste_mode:
        paste_mode_context_manager = bracketed_paste_mode(stdout)
    else:
        paste_mode_context_manager = nullcontext()

    with paste_mode_context_manager:
        while True:
            exit_sequence = raw_serial_terminal(
                conn=conn,
                stdin=stdin,
                stdout=stdout,
                exit_on=(
                    ([BRACKETED_PASTE_BEGIN] if automatic_paste_mode else [])
                    + (["\x0c"] if emulate_ctrl_l else [])
                    + exit_on
                ),
            )

            if exit_sequence == BRACKETED_PASTE_BEGIN:
                handle_bracketed_paste(
                    conn=conn,
                    stdin=stdin,
                    stdout=stdout,
                )
            elif exit_sequence == "\x0c":  # Ctrl+L
                handle_ctrl_l_emulation(conn, stdout)
            else:
                return exit_sequence



"""
Enumerate files and directories on a MicroPython device.
"""

import sys

from argparse import ArgumentParser, Namespace

from upyt.connection import Connection
from upyt.upy_terminal import interrupt_and_enter_repl
from upyt.upy_fs import upy_filesystem
from upyt.cli.hybrid_filesystem_api import HybridFilesystemAPI


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "path",
        nargs="?",
        default=":/",
        help="""
            The path to enumerate on the device. Defaults to :/. Prefix with
            ':' for device paths.
        """,
    )

    parser.add_argument(
        "--long",
        "-l",
        action="store_true",
        help="""
            If given, show additional details about each file.
        """,
    )


def main(args: Namespace):
    with Connection.from_specification(args.device) as conn:
        interrupt_and_enter_repl(conn)
        with upy_filesystem(conn) as fs:
            if not args.path.startswith(":"):
                print(
                    "warning: path not on device (i.e. starting with ':')",
                    file=sys.stderr,
                )
            hfs = HybridFilesystemAPI(fs)

            directories, files = hfs.ls(args.path)
            entries = sorted(
                (entry, typ)
                for lst, typ in [(directories, "d"), (files, "-")]
                for entry in lst
            )

            for name, typ in entries:
                if typ == "d":
                    name = f"{name}/"

                if args.long:
                    if typ != "d":
                        size = hfs.file_len(f"{args.path}/{name}")
                    else:
                        size = 0
                    print(f"{size:8d} {name}")
                else:
                    print(name)



"""
Low-level utilities which interact with the MicroPython REPL.
"""

import random

from contextlib import contextmanager

import struct

from upyt.connection import Connection


class MicroPythonReplError(Exception):
    """
    Base exception thrown when the MicroPython REPL produces unexpected output.

    All exceptions include any unexpected output from the connection as their
    first argument.
    """


class NoReplError(MicroPythonReplError):
    """
    Thrown when no REPL prompt (>>>) is received when expected.
    """


class CouldntEnterRawPasteModeError(MicroPythonReplError):
    """
    Thrown if :py:func:`paste_exec` is unable to enter paste mode.
    """


class RawPasteModeNotSupportedError(MicroPythonReplError):
    """
    Thrown if the connected board is running a MicroPython version too old to
    support raw paste mode.
    """


class SomeCodeNotSentError(MicroPythonReplError):
    """
    Thrown the MicroPython board fails to accept all of the code provided to
    :py:func:`raw_paste_exec`.

    The first argument is th exception output (as raw bytes) and the second
    argument is the ordinary output (also as raw bytes, and likely empty). The
    third argument is the input which was never sent.
    """


def expect(conn: Connection, value: bytes) -> bytes:
    """
    Read from a connection, checking that exactly the expected value is read.
    If a timeout occurs, raises MicroPythonReplError.

    Returns the bytes actually read.
    """
    actual = conn.read(len(value))
    if actual != value:
        raise MicroPythonReplError(actual)
    return actual


def expect_endswith(conn: Connection, value: bytes) -> bytes:
    """
    Read from a connection until the supplied value is read. If a timeout
    occurs before this, raises MicroPythonReplError.

    Returns all of the bytes read.
    """
    actual = conn.read_until(value)
    if not actual.endswith(value):
        raise MicroPythonReplError(actual)
    return actual


def interrupt_and_enter_repl(
    conn: Connection,
    num_attempts: int = 2,
) -> bytes:
    """
    Attempt to perform a keyboard interrupt to get to a REPL.

    Raises NoReplError if the REPL is not reached.

    Returns all ignored terminal output prior to the final newline and prompt
    ('>>>'), e.g.  KeyboardInterrupt exception trace-backs).

    Makes num_attempts attempts to interrupt the running process, waiting
    conn.timeout seconds between each attempt. Multiple attempts are necessary if,
    for example, an exception handler also blocks.
    """
    # Flush any pending input (e.g. previous command outputs or prompts)
    unmatched_output = conn.read_buffered()

    # Attempt to reach a prompt follwing a keyboard interrupt
    for attempt in range(num_attempts):
        conn.write(b"\x03")

        # Wait for the first hint hint of a prompt
        prompt = b"\r\n>>> "
        try:
            unmatched_output += expect_endswith(conn, prompt)[: -len(prompt)]
        except MicroPythonReplError as e:
            # No sign of a prompt, try interrupting again as the exception
            # handler may need interrupting
            unmatched_output += e.args[0]
            continue

        # To make sure what we're seeing is not just an old prompt hanging
        # about in the buffer, produce some unique output
        random_number = random.randint(0x10, 0xFFFFFF)
        conn.write(b"0x%x\r" % random_number)
        expected_response = (
            b"0x%x\r\n"  # Echo
            b"%d\r\n"  # Decimal representation
            b">>> "  # Fresh prompt
        ) % (random_number, random_number)
        try:
            unmatched_output += expect_endswith(conn, expected_response)[
                : -len(expected_response)
            ]
            # Success! We're in sync and sat waiting at a prompt!
            return unmatched_output
        except MicroPythonReplError as e:
            # Fail: we didn't see our reply, lets try agian
            unmatched_output += e.args[0]
            continue

    # Ran out of retries
    raise NoReplError(unmatched_output)


def paste_exec(conn: Connection, code: str, batch_size: int = 32) -> None:
    """
    (Attempt to) use paste-mode to execute a block of code.

    This function expects to be execued when the REPL is sat at a fresh prompt
    (e.g. following a call to :py:func:`interrupt_and_enter_repl`).

    The only reason to use this function is when you want to run some code
    which the user has pasted and which may run for an indeterminate time which
    the user can deal with waiting for or giving up on.

    In almost all other circumstances you should use :py:meth:`raw_paste_exec`
    to execute arbitrary code on the device.

    Parameters
    ==========
    conn : Connection
    code : str
    batch_size : int
        The number of bytes to write to the device at a time before waiting for
        the terminal echo back. (This is a crude form of flow control).
    """
    # Enter paste mode
    conn.write(b"\x05")  # (Ctrl+E)
    expect_endswith(conn, b"paste mode; Ctrl-C to cancel, Ctrl-D to finish\r\n=== ")

    # Write out batch_size chars at a time
    code_bytes = code.encode("utf-8")
    while code_bytes:
        batch = code_bytes[:batch_size]
        code_bytes = code_bytes[batch_size:]

        conn.write(batch)
        # Wait for echo back (including paste mode prompts)
        expect(conn, batch.replace(b"\r", b"\r\n=== "))

    # Exit paste mode and execute
    conn.write(b"\x04")  # (Ctrl+D)
    expect(conn, b"\r\n")


@contextmanager
def raw_mode(conn: Connection):
    """
    Context manager which enters raw mode on entry and leaves raw mode again on
    exit.
    """
    # Enter raw mode
    conn.write(b"\x01")  # (Ctrl+A)
    expect_endswith(conn, b"raw REPL; CTRL-B to exit\r\n>")

    try:
        yield
    finally:
        conn.write(b"\x02")  # (Ctrl+B)  Exit raw mode
        expect_endswith(conn, b"\r\n>>> ")


def raw_paste_exec(conn: Connection, code: str) -> tuple[str, str]:
    """
    Execute the supplied code (via raw paste mode).

    Must be in raw mode (see :py:func:`raw_mode`) when calling this function.

    The supplied code must not contain, or print a 0x04 character (Ctrl+D)
    otherwise the raw paste mode protocol will become out-of-sync and undefined
    behaviour will ensue.

    Returns two strings: the output of the executed code and any exceptions
    produced.

    The executed command must complete and have sent all of its output within
    the connection's timeout, otherwise an error will occur.
    """
    code_utf8 = code.encode("utf-8")
    if b"\x04" in code_utf8:
        raise ValueError("Cannot eval strings containing ASCII 0x04 (ctrl+D)")

    # Enter raw paste mode
    conn.write(b"\x05" b"A" b"\x01")  # (Ctrl+E / ENQ)  # (Ctrl+A)  # (Ctrl+A)
    response = conn.read(2)
    if response != b"R\x01":
        raise RawPasteModeNotSupportedError(response)

    window_size_increment = struct.unpack("<H", conn.read(2))[0]
    window_size = window_size_increment

    # Send all of the code and wait for the window to re-open again
    #
    # NB: We need to wait for the window to reopen before sending the
    # end-of-code if we happened to use up the whole window with our final
    # block of code, otherwise the end-of-code might overrun the buffer!
    while code_utf8 or window_size == 0:
        if window_size == 0:
            response = conn.read(1)
            if response == b"\x01":
                window_size += window_size_increment
            elif response == b"\x04":
                # Device doesn't want any more data
                break
            else:
                raise MicroPythonReplError(response)

        # Send (up-to) the maxmimum allowed window worth of data
        written = conn.write(code_utf8[:window_size])
        code_utf8 = code_utf8[written:]
        window_size -= written

    # End transmission
    conn.write(b"\x04")  # (Ctrl+D)
    while True:
        response = conn.read(1)
        if response == b"\x01":
            # A window size increment, but we don't care about those anymore
            continue
        elif response == b"\x04":
            # End of transmission was acknowledged
            break
        else:
            raise MicroPythonReplError(response)

    # Read the response
    code_output = expect_endswith(conn, b"\x04")[:-1]
    exception_output = expect_endswith(conn, b"\x04")[:-1]
    expect_endswith(conn, b">")  # Should return to raw-repl shell

    if not code_utf8:
        # Success: All bytes were accepted!
        return (code_output.decode("utf-8"), exception_output.decode("utf-8"))
    else:
        # Failure: Some bytes not received
        raise SomeCodeNotSentError(exception_output, code_output, code_utf8)


def soft_reset_directly_into_repl(conn: Connection) -> str:
    """
    Interupt any running process and perform a soft reset such that the device
    boots directly into the REPL without executing main.py.

    Returns any output produced by boot.py as a string. Output prior to this is
    discarded.
    """
    interrupt_and_enter_repl(conn)

    # NB: When reset in raw mode, main.py is not executed
    with raw_mode(conn):
        # Perform a reset
        conn.write(b"\x04")  # (Ctrl + D)

        # Response from Raw REPL
        expect(
            conn,
            (b"OK\r\n" b"MPY: soft reboot\r\n"),  # From Raw REPL  # Boot message
        )

        # Ignore all output from boot.py
        raw_repl_entry_message = b"raw REPL; CTRL-B to exit\r\n>"
        boot_py_output = expect_endswith(conn, raw_repl_entry_message)[
            : -len(raw_repl_entry_message)
        ]

    return boot_py_output.decode("utf-8")



"""
This module implements a wrapper around :py:class:`upyt.upy_fs.FilesystemAPI`
which accesses files on the host when they do not start with a colon (':') and
on the device when they do.
"""

from upyt.upy_fs import FilesystemAPI, PathType

import os
from pathlib import Path
from shutil import rmtree


class HybridFilesystemAPI:
    def __init__(self, fs: FilesystemAPI) -> None:
        self._fs = fs

    def get_type(self, path: str) -> PathType:
        if path.startswith(":"):
            return self._fs.get_type(path[1:])
        else:
            if Path(path).exists():
                if Path(path).is_dir():
                    return PathType.dir
                else:
                    return PathType.file
            else:
                return PathType.absent

    def mkdir(self, path: str, parents: bool = False, exist_ok: bool = False) -> None:
        if path.startswith(":"):
            self._fs.mkdir(path[1:], exist_ok=exist_ok, parents=parents)
        else:
            Path(path).mkdir(exist_ok=exist_ok, parents=parents)

    def remove_recursive(self, path: str) -> None:
        if path.startswith(":"):
            self._fs.remove_recursive(path[1:])
        else:
            if Path(path).is_file():
                Path(path).unlink()
            else:
                rmtree(Path(path))

    def ls(self, path: str) -> tuple[list[str], list[str]]:
        if path.startswith(":"):
            return self._fs.ls(path[1:])
        else:
            dirs = []
            files = []
            for entry in Path(path).iterdir():
                if entry.is_dir():
                    dirs.append(entry.name)
                else:
                    files.append(entry.name)
            return (dirs, files)

    def rename(self, old_path: str, new_path: str) -> None:
        assert old_path.startswith(":") is new_path.startswith(":")

        if old_path.startswith(":"):
            self._fs.rename(old_path[1:], new_path[1:])
        else:
            Path(old_path).replace(new_path)

    def write_file(self, path: str, content: bytes) -> None:
        if path.startswith(":"):
            self._fs.write_file(path[1:], content)
        else:
            Path(path).write_bytes(content)

    def read_file(self, path: str) -> bytes:
        if path.startswith(":"):
            return self._fs.read_file(path[1:])
        else:
            return Path(path).read_bytes()

    def file_len(self, path: str) -> int:
        if path.startswith(":"):
            return self._fs.file_len(path[1:])
        else:
            return Path(path).stat().st_size

    def sync(self) -> None:
        self._fs.sync()
        os.sync()

    def update_file(self, *_, **__) -> None:
        raise NotImplementedError()



"""
Soft-reset a MicroPython device.
"""

import sys

from argparse import ArgumentParser, Namespace

from upyt.connection import Connection
from upyt.upy_repl import (
    expect,
    interrupt_and_enter_repl,
    soft_reset_directly_into_repl,
)


def add_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--repl",
        "-r",
        action="store_true",
        default=False,
        help="""
            If given, force the device to reset into a REPL, without running
            main.py (if it exists).
        """,
    )


def main(args: Namespace):
    with Connection.from_specification(args.device) as conn:
        interrupt_and_enter_repl(conn)
        if args.repl:
            soft_reset_directly_into_repl(conn)
        else:
            conn.write(b"\x04")  # ctrl+d
            expect(conn, b"\r\n")



from subprocess import run

from pathlib import Path

import upyt


def test_with_mypy() -> None:
    paths = [
        str(Path(upyt.__file__).parent),  # upyt dir
        str(Path(__file__).parent),  # tests dir
    ]

    run(["mypy"] + paths, check=True)



import pytest

from typing import Any
from pathlib import Path

from upyt.upy_fs import FilesystemAPI

from trees import (
    write_device_tree,
    read_device_tree,
    write_local_tree,
    read_local_tree,
)


@pytest.mark.parametrize(
    "tree",
    [
        # Empty
        {},
        # Files
        {"a.txt": b"I am a"},
        {"a.txt": b"I am a", "b.txt": b"I am b"},
        # Directories
        {"a": {}},
        {"a": {"b": {}}},
    ],
)
def test_read_write_device_tree_roundtrip(
    fs: FilesystemAPI,
    dev_tmpdir: str,
    tree: dict[str, Any] | bytes,
) -> None:
    write_device_tree(fs, dev_tmpdir, tree)
    assert read_device_tree(fs, dev_tmpdir) == tree


@pytest.mark.parametrize(
    "tree",
    [
        # Empty
        {},
        # Files
        {"a.txt": b"I am a"},
        {"a.txt": b"I am a", "b.txt": b"I am b"},
        # Directories
        {"a": {}},
        {"a": {"b": {}}},
    ],
)
def test_read_write_local_tree_roundtrip(
    tmp_path: Path,
    tree: dict[str, Any] | bytes,
) -> None:
    write_local_tree(tmp_path, tree)
    assert read_local_tree(tmp_path) == tree
